"""Model fit in maximum likelihood or Bayesian way."""

from __future__ import annotations

import time
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import TYPE_CHECKING

import jax
import jax.numpy as jnp
import nautilus
import numpy as np
import optimistix as optx
import ultranest
from iminuit import Minuit
from jax.experimental.mesh_utils import create_device_mesh
from jax.experimental.shard_map import shard_map
from jax.sharding import Mesh, PartitionSpec
from numpyro.infer import AIES, MCMC, NUTS, init_to_value

from elisa.data.base import FixedData, ObservationData
from elisa.infer.helper import Helper, get_helper
from elisa.infer.likelihood import _STATISTIC_OPTIONS
from elisa.infer.nested_sampling import NestedSampler, reparam_loglike
from elisa.infer.results import MLEResult, PosteriorResult
from elisa.models.model import Model, get_model_info
from elisa.util.misc import add_suffix, build_namespace, make_pretty_table

if TYPE_CHECKING:
    from typing import Any, Callable, Literal

    from prettytable import PrettyTable

    from elisa.infer.likelihood import Statistic
    from elisa.models.model import ModelInfo
    from elisa.util.typing import ArrayLike, JAXArray, JAXFloat


class Fit(ABC):
    """Abstract base class for model fitting.

    Parameters
    ----------
    data : Data or sequence of Data
        The observation data.
    model : Model or sequence of Model
        The model used to fit the data.
    stat : {'chi2', 'cstat', 'pstat', 'pgstat', 'wstat'} or sequence, optional
        The likelihood option for the data and model. Available likelihood
        options are:

            * ``'chi2'``: Gaussian data
            * ``'cstat'``: Poisson data
            * ``'pstat'``: Poisson data with known background
            * ``'pgstat'``: Poisson data with Gaussian background
            * ``'wstat'``: Poisson data with Poisson background

        The default is None, which means automatically choosing the suitable
        likelihood options for the datasets and models.
    seed : int, optional
        Seed of random number generator used for fit. The default is 42.
    """

    # TODO:
    #  - fit multiple sources to one dataset (with multiple responses)
    #  - fit data background given response and model

    _lm: Callable[[JAXArray], JAXArray] | None = None
    _ns: NestedSampler | None = None

    def __init__(
        self,
        data: ObservationData | Sequence[ObservationData],
        model: Model | Sequence[Model],
        stat: Statistic | Sequence[Statistic] | None = None,
        seed: int = 42,
    ):
        inputs = self._parse_input(data, model, stat)
        data: list[FixedData] = inputs[0]
        models: list[Model] = inputs[1]
        stats: list[Statistic] = inputs[2]

        # if a component is not fit with all datasets,
        # add names of data sets to be fit with it as its name/latex suffix
        data_names = [d.name for d in data]
        data_to_cid = {n: m._comps_id for n, m in zip(data_names, models)}
        cid_to_comp = {c._id: c for m in models for c in m._comps}
        cid = list(cid_to_comp.keys())
        comps = list(cid_to_comp.values())
        cid_to_data_suffix = {
            i: '+'.join(i for i in data_names if i not in names)  # keep order
            if (names := [n for n in data_names if i not in data_to_cid[n]])
            else ''
            for i in cid
        }
        data_suffix = list(cid_to_data_suffix.values())
        cname = [comp.name for comp in comps]
        name_with_data_suffix = list(map(''.join, zip(cname, data_suffix)))
        num_suffix = build_namespace(name_with_data_suffix)['suffix_num']
        cname = add_suffix(cname, num_suffix, True, True)
        cname = add_suffix(cname, data_suffix, False, True)
        cid_to_name = dict(zip(cid, cname))
        latex = [comp.latex for comp in comps]
        latex = add_suffix(latex, num_suffix, True, latex=True)
        latex = add_suffix(latex, data_suffix, False, latex=True, mathrm=True)
        cid_to_latex = dict(zip(cid, latex))

        # get model info
        self._model_info: ModelInfo = get_model_info(
            comps, cid_to_name, cid_to_latex
        )

        # first filter out duplicated models then compile the remaining models,
        # this is intended to avoid re-compilation of the same model
        models_id = [id(m) for m in models]
        mid_to_model = dict(zip(models_id, models))
        compiled_model = {
            mid: m.compile(model_info=self._model_info)
            for mid, m in mid_to_model.items()
        }
        data_to_mid = dict(zip(data_names, models_id))
        self._model = {
            name: compiled_model[mid] for name, mid in data_to_mid.items()
        }

        # store data, stat, seed
        self._data: dict[str, FixedData] = dict(zip(data_names, data))
        self._stat: dict[str, Statistic] = dict(zip(data_names, stats))
        self._seed: int = int(seed)

        # make model information table
        self._make_info_table()

        self.__helper: Helper | None = None

    def _optimize_lm(
        self,
        unconstr_init: JAXArray,
        max_steps: int = 131072,
        throw: bool = True,
        verbose: bool = False,
    ) -> tuple[JAXArray, JAXFloat]:
        """Search MLE by Levenberg-Marquardt algorithm of :mod:`optimistix`."""

        if verbose:
            verbose = frozenset({'step', 'loss'})
        else:
            verbose = frozenset()

        if self._lm is None:
            lm_solver = optx.LevenbergMarquardt(
                rtol=0.0, atol=1e-6, verbose=verbose
            )
            residual = jax.jit(lambda x, aux: self._helper.residual(x))

            def lm(init):
                res = optx.least_squares(
                    fn=residual,
                    solver=lm_solver,
                    y0=init,
                    max_steps=max_steps,
                    throw=throw,
                )
                return res.value, jnp.linalg.norm(res.state.f_info.grad)

            self._lm = jax.jit(lm)

        return self._lm(jnp.asarray(unconstr_init, float))

    def _optimize_ns(self, max_steps=131072, verbose=False) -> JAXArray:
        """Search MLE using nested sampling of :mod:`jaxns`."""
        if self._ns is None:
            self._ns = NestedSampler(
                self._helper.numpyro_model,
                constructor_kwargs={
                    'max_samples': max_steps,
                    'parameter_estimation': True,
                    'verbose': verbose,
                },
            )
            t0 = time.time()
            print('Start searching MLE...')
            self._ns.run(rng_key=jax.random.PRNGKey(self._helper.seed['mcmc']))
            print(f'Search cost {time.time() - t0:.2f} s')

        ns = self._ns

        samples = ns._results.samples
        loglike = [
            samples[f'{i}_loglike'].sum(axis=-1) for i in self._data.keys()
        ]
        mle_idx = np.sum(loglike, axis=0).argmax()
        mle = jax.tree_map(lambda s: s[mle_idx], samples)
        mle = {i: mle[i] for i in self._helper.params_names['free']}
        return self._helper.constr_dic_to_unconstr_arr(mle)

    @property
    def _helper(self) -> Helper:
        if self.__helper is None:
            self.__helper = get_helper(self)

        return self.__helper

    def summary(self, file=None) -> None:
        """Print the summary of fitting setup.

        Parameters
        ----------
        file : file-like
            An object with a ``write(string)`` method. This is passed to
            :py:func:`print`.
        """
        print(repr(self), file=file)

    @property
    @abstractmethod
    def _tab_config(self) -> tuple[str, frozenset[str]]:
        """Model information table's title and excluded table fields."""
        pass

    def __repr__(self) -> str:
        return (
            f'\n{self._tab_config[0]}\n\n'
            f'{self._tab_likelihood.get_string()}\n\n'
            f'{self._tab_params.get_string()}\n'
        )

    def _repr_html_(self) -> str:
        """The repr in Jupyter notebook environment."""
        return (
            f'<details open><summary><b>{self._tab_config[0]}</b></summary>'
            f'<br/>{self._tab_likelihood.get_html_string(format=True)}'
            f'<br/>{self._tab_params.get_html_string(format=True)}'
            '</details>'
        )

    def _make_info_table(self):
        fields = ('Data', 'Model', 'Statistic')
        rows = tuple(
            zip(
                self._data,
                (m.name for m in self._model.values()),
                self._stat.values(),
            )
        )
        self._tab_likelihood: PrettyTable = make_pretty_table(fields, rows)

        fields = ('No.', 'Component', 'Parameter', 'Value', 'Bound', 'Prior')
        mask = np.isin(fields, tuple(self._tab_config[1]))
        fields = np.array(fields)[~mask].tolist()
        rows = np.array(self._model_info.info)[:, ~mask].tolist()
        self._tab_params: PrettyTable = make_pretty_table(fields, rows)

    @staticmethod
    def _parse_input(
        data: ObservationData | Sequence[ObservationData],
        model: Model | Sequence[Model],
        stat: Statistic | Sequence[Statistic] | None,
    ) -> tuple[list[FixedData], list[Model], list[Statistic]]:
        """Check if data, model, and stat are correct and return lists."""

        # ====================== some helper functions ========================
        def get_list(
            inputs: Any, name: str, expect_type, type_name: str
        ) -> list:
            """Check the model/data/stat, and return a list."""
            if isinstance(inputs, expect_type):
                input_list = [inputs]
            elif isinstance(inputs, Sequence):
                if not inputs:
                    raise ValueError(f'{name} list is empty')
                if not all(isinstance(i, expect_type) for i in inputs):
                    raise ValueError(f'all {name} must be a valid {type_name}')
                input_list = list(inputs)
            else:
                raise ValueError(f'got wrong type {type(inputs)} for {name}')
            return input_list

        def get_stat(d: FixedData) -> Statistic:
            """Get the default stat for the data."""
            # 'pstat' is used only when specified explicitly by user
            if d.spec_poisson:
                if d.has_back:
                    if d.back_poisson:
                        return 'wstat'
                    return 'pgstat'
                else:
                    return 'cstat'
            else:
                return 'chi2'

        def check_stat(d: FixedData, s: Statistic):
            """Check if data type and likelihood are matched."""
            name = d.name
            if not d.spec_poisson and s != 'chi2':
                msg = f'{name} is Gaussian data, use stat "chi2" instead'
                raise ValueError(msg)

            if s == 'chi2' and np.any(d.net_errors == 0.0):
                raise ValueError(
                    f'"chi2" is not valid for {name} data, which has zero '
                    'uncertainties; grouping the data may fix this error'
                )

            elif s == 'cstat' and d.has_back:
                back = 'Poisson' if d.back_poisson else 'Gaussian'
                stat1 = 'W' if d.back_poisson else 'PG'
                stat2 = 'w' if d.back_poisson else 'pg'
                msg = 'C-statistic (cstat) is not valid for Poisson data '
                msg += f'with {back} background, use {stat1}-statistic'
                msg += f'({stat2}stat) for {name} instead'
                raise ValueError(msg)

            elif s == 'pstat' and not d.has_back:
                msg = f'P-statistic (pstat) is not valid for {name}, which '
                msg += 'requires background file, use C-statistic (cstat) '
                msg += 'instead'
                raise ValueError(msg)

            elif s == 'pgstat':
                if not d.has_back:
                    msg = f'PG-statistic is not valid for {name}, which '
                    msg += 'requires Gaussian background data, '
                    msg += 'use C-statistic instead (cstat)'
                    raise ValueError(msg)

                if np.any(d.back_errors == 0.0):
                    raise ValueError(
                        f'PG-statistic is not valid for {name} data, '
                        'which has zero background uncertainties; '
                        'grouping the data may fix this error'
                    )

            elif s == 'wstat' and not (d.has_back and d.back_poisson):
                msg = f'W-statistic is not valid for {name}, which requires '
                msg += 'Poisson background data, use C-statistic (cstat) '
                msg += 'instead'
                raise ValueError(msg)

        # ====================== some helper functions ========================

        # get data
        data_list: list[FixedData] = [
            d.get_fixed_data()
            for d in get_list(data, 'data', ObservationData, 'Data')
        ]

        # check if data are used multiple times
        if len(list(map(id, data_list))) != len(data_list):
            count = {d: data_list.count(d) for d in set(data_list)}
            raise ValueError(
                'data cannot be used multiple times: '
                + ', '.join(
                    f'{k.name} ({v})' for k, v in count.items() if v > 1
                )
            )

        # check if data name is unique
        name_list = [d.name for d in data_list]
        if len(set(name_list)) != len(data_list):
            raise ValueError(
                f'data names are not unique: {", ".join(set(name_list))}, '
                "please give a unique name in Data(..., name='NAME'), "
                "or set data.name='NAME'"
            )

        # get model
        model_list: list[Model] = get_list(model, 'model', Model, 'Model')

        # check if the model type is additive
        flag = [i.type == 'add' for i in model_list]
        if not all(flag):
            err = (j for i, j in enumerate(model_list) if not flag[i])
            err = ', '.join(f"'{i}'" for i in err)
            msg = f'got models which are not additive type: {err}'
            raise TypeError(msg)

        # get stat
        stat_list: list[Statistic]
        if stat is None:
            stat_list: list[Statistic] = [get_stat(d) for d in data_list]
        else:
            stat_list: list[Statistic] = get_list(stat, 'stat', str, 'str')

            # check the stat option
            flag = [i in _STATISTIC_OPTIONS for i in stat_list]
            if not all(flag):
                err = ', '.join(
                    f"'{j}'" for i, j in enumerate(stat_list) if not flag[i]
                )
                supported = ', '.join(f"'{i}'" for i in _STATISTIC_OPTIONS)
                msg = f'unexpected stat: {err}; supported are {supported}'
                raise ValueError(msg)

        nd = len(data_list)
        nm = len(model_list)
        ns = len(stat_list)

        # check model number
        if nm == 1:
            model_list *= nd
        elif nm != nd:
            msg = f'number of model ({nm}) and data ({nd}) are not matched'
            raise ValueError(msg)

        # check stat number
        if ns == 1:
            stat_list *= nd
        elif ns != nd:
            msg = f'number of data ({nd}) and stat ({ns}) are not matched'
            raise ValueError(msg)

        # check if correctly using stat
        for d, s in zip(data_list, stat_list):
            check_stat(d, s)

        return data_list, model_list, stat_list


class MaxLikeFit(Fit):
    _tab_config = ('Maximum Likelihood Fit', frozenset({'Prior'}))

    def _optimize_minuit(
        self,
        unconstr_init: JAXArray,
        ncall: int | None = None,
        throw: bool = True,
        verbose: int | bool = False,
    ) -> Minuit:
        """Search MLE using Minuit algorithm of :mod:`iminuit`."""
        deviance = jax.jit(self._helper.deviance_total)
        deviance.ndata = self._helper.ndata['total']
        minuit = Minuit(
            deviance,
            np.array(unconstr_init),
            grad=jax.jit(jax.grad(deviance)),
            name=self._helper.params_names['free'],
        )

        if throw:
            minuit.throw_nan = True

        minuit.print_level = int(verbose)

        # TODO: test if simplex can be used to "polish" the initial guess
        minuit.strategy = 2
        minuit.migrad(ncall=ncall, iterate=10)

        return minuit

    def mle(
        self,
        init: ArrayLike | dict | None = None,
        method: Literal['minuit', 'lm', 'ns'] = 'minuit',
        max_steps: int = None,
        throw: bool = True,
        verbose: int | bool = False,
    ) -> MLEResult:
        """Search Maximum Likelihood Estimation (MLE) for the model.

        Parameters
        ----------
        init : array_like or dict, optional
            Initial guess for the maximum likelihood estimation. The default is
            as parameters' default values.
        method : {'minuit', 'lm', 'ns'}, optional
            Optimization algorithm used to find the MLE.
            Available options are:

                * ``'minuit'``: Migrad algorithm of :mod:`iminuit`.
                * ``'lm'``: Levenberg-Marquardt algorithm of :mod:`optimistix`.
                * ``'ns'``: Nested sampling of :mod:`jaxns`. This option first
                  search MLE globally, then polish it with local minimization.

            The default is 'minuit'.

        Other Parameters
        ----------------
        max_steps : int, optional
            The maximum number of steps the solver can take. The default is
            131072.
        throw : bool, optional
            Whether to report any failures of the solver. Defaults to True.
        verbose : int or bool, optional
            Whether to print fit progress information. The default is False.

        Returns
        -------
        MLEResult
            The MLE result.
        """
        if isinstance(init, (np.ndarray, jax.Array, Sequence)):
            init_unconstr = self._helper.constr_arr_to_unconstr_arr(init)
        elif isinstance(init, dict):
            init_unconstr = self._helper.constr_dic_to_unconstr_arr(init)
        elif init is None:
            init_unconstr = self._helper.free_default['unconstr_arr']
        else:
            raise TypeError('params must be a array, sequence, or mapping')

        max_steps = 131072 if max_steps is None else int(max_steps)

        if method == 'lm':  # use Levenberg-Marquardt algorithm to find MLE
            init_unconstr, _ = self._optimize_lm(
                init_unconstr, max_steps, throw, bool(verbose)
            )
        elif method == 'ns':  # use nested sampling to find MLE
            init_unconstr = self._optimize_ns(max_steps, verbose)
        else:
            if method != 'minuit':
                raise ValueError(f'unsupported optimization method {method}')

        minuit = self._optimize_minuit(
            init_unconstr, max_steps, throw, verbose
        )

        return MLEResult(minuit, self._helper)


class BayesFit(Fit):
    _tab_config = ('Bayesian Fit', frozenset({'Bound'}))

    def nuts(
        self,
        warmup=2000,
        samples=20000,
        chains: int | None = None,
        init: dict[str, float] | None = None,
        chain_method: str = 'parallel',
        progress: bool = True,
        **nuts_kwargs: dict,
    ) -> PosteriorResult:
        """Run the No-U-Turn Sampler of :mod:`numpyro`.

        .. note::
            If the chains are not converged well, see ref [1]_ for more
            information on how to finetune the NUTS sampler.

        Parameters
        ----------
        warmup : int, optional
            Number of warmup steps.
        samples : int, optional
            Number of samples to generate from each chain.
        chains : int, optional
            Number of MCMC chains to run. If there are not enough devices
            available, chains will run in sequence. Defaults to the number of
            ``jax.local_device_count()``.
        init : dict, optional
            Initial parameter for sampler to start from.
        chain_method : str, optional
            The chain method passed to :class:`numpyro.infer.MCMC`.
        progress : bool, optional
            Whether to show progress bar during sampling. The default is True.
        **nuts_kwargs : dict
            Extra parameters passed to :class:`numpyro.infer.NUTS`.

        Returns
        -------
        PosteriorResult
            The posterior sampling result.

        References
        ----------
        .. [1] NumPyro tutorial: `Bad posterior geometry and how to deal with
               it <https://num.pyro.ai/en/stable/tutorials/bad_posterior_geometry.html>`__
        """
        device_count = jax.local_device_count()

        if chains is None:
            chains = device_count
        else:
            chains = int(chains)

        samples = int(samples)

        # the total samples number should be multiple of the device number
        if chains * samples % device_count != 0:
            samples += device_count - samples % device_count

        # TODO: option to let sampler starting from MLE
        if init is None:
            init = self._helper.free_default['constr_dic']
        else:
            init = self._helper.free_default['constr_dic'] | dict(init)

        default_nuts_kwargs = {
            'dense_mass': True,
            'target_accept_prob': 0.8,
            'max_tree_depth': 10,
        }
        nuts_kwargs = default_nuts_kwargs | nuts_kwargs
        nuts_kwargs['model'] = self._helper.numpyro_model
        nuts_kwargs['init_strategy'] = init_to_value(values=init)

        sampler = MCMC(
            NUTS(**nuts_kwargs),
            num_warmup=warmup,
            num_samples=samples,
            num_chains=chains,
            chain_method=chain_method,
            progress_bar=progress,
        )

        sampler.run(
            rng_key=jax.random.PRNGKey(self._helper.seed['mcmc']),
            extra_fields=('energy', 'num_steps'),
        )
        return PosteriorResult(sampler, self._helper, self)

    def jaxns(
        self,
        max_samples: int = 131072,
        num_live_points: int | None = None,
        s: int | None = None,
        k: int | None = None,
        c: int | None = None,
        num_parallel_workers: int = 1,
        difficult_model: bool = False,
        parameter_estimation: bool = False,
        verbose: bool = False,
        term_cond: dict | None = None,
        **ns_kwargs: dict,
    ) -> PosteriorResult:
        """Run the Nested Sampler of :mod:`jaxns`.

        .. note::
            Parameters `s`, `k`, and `c` are defined in the paper [1]_.
            For more information of the sampler parameters, see ref [1]_ [2]_.

        Parameters
        ----------
        max_samples : int, optional
            Maximum number of posterior samples. The default is 131072.
        num_live_points : int, optional
            Approximate number of live points. The default is `c` * (`k` + 1).
        s : int, optional
            Number of slices per dimension. The default is 5.
        k : int, optional
            Number of phantom samples. The default is 0.
        c : int, optional
            Number of parallel Markov chains. The default is 30 * `D`, where
            `D` is the dimension of model parameters. It takes effect only
            for num_live_points=None.
        num_parallel_workers : int, optional
            Parallel workers number. The default is 1.
        difficult_model : bool, optional
            If True, uses more robust default settings (`s` = 10 and
            `c` = 50 * `D`). It takes effect only for `num_live_points` = None,
            `s` = None or `c` = None. Defaults to False.
        parameter_estimation : bool, optional
            If True, uses more robust default settings for parameter estimation
            (`k` = `D`). It takes effect only for `k` = None.
            Defaults to False.
        verbose : bool, optional
            Print progress information. The default is False.
        term_cond : dict, optional
            Termination conditions for the sampling. The default is as in
            :class:`jaxns.TermCondition`.
        **ns_kwargs : dict
            Extra parameters passes to :class:`jaxns.DefaultNestedSampler`.

        Returns
        -------
        PosteriorResult
            The posterior sampling result.

        References
        ----------
        .. [1] `Phantom-Powered Nested Sampling <https://arxiv.org/abs/2312.11330>`__
        .. [2] `JAXNS API doc <https://jaxns.readthedocs.io/en/latest/api/jaxns/index.html#jaxns.DefaultNestedSampler>`__
        """
        num_parallel_workers = int(num_parallel_workers)

        constructor_kwargs = {
            'max_samples': max_samples,
            'num_live_points': num_live_points,
            's': s,
            'k': k,
            'c': c,
            'num_parallel_workers': num_parallel_workers,
            'difficult_model': difficult_model,
            'parameter_estimation': parameter_estimation,
            'verbose': verbose,
        }
        constructor_kwargs |= ns_kwargs

        termination_kwargs = {'dlogZ': 1e-4}
        if term_cond is not None:
            termination_kwargs |= dict(term_cond)

        sampler = NestedSampler(
            self._helper.numpyro_model,
            constructor_kwargs=constructor_kwargs,
            termination_kwargs=termination_kwargs,
        )

        print('Start nested sampling...')
        t0 = time.time()
        sampler.run(rng_key=jax.random.PRNGKey(self._helper.seed['mcmc']))
        print(f'Sampling cost {time.time() - t0:.2f} s')
        return PosteriorResult(sampler, self._helper, self)

    def ultranest(
        self,
        ess: int = 10000,
        ignore_nan: bool = True,
        *,
        constructor_kwargs: dict | None = None,
        termination_kwargs: dict | None = None,
        read_file: dict | None = None,
    ) -> PosteriorResult:
        """Run the Nested Sampler of :mod:`ultranest`.

        Parameters
        ----------
        ess : int, optional
            The desired effective sample size.
        ignore_nan : bool, optional
            Whether to transform a NaN log probability to a large negative
            number (-1e300). The default is True.

            .. warning ::
                Setting ``ignore_nan=True`` may fail to spot potential issues
                with model computation.
        constructor_kwargs : dict, optional
            Extra parameters passed to
            :class:`ultranest.ReactiveNestedSampler`.
        termination_kwargs : dict, optional
            Extra parameters passed to
            :class:`ultranest.ReactiveNestedSampler.run()`.
        read_file : dict, optional
            Read the log file from a previous run. The dictionary should
            contain the log directory and other optional parameters. It
            should be noted that when providing this keyword argument, the
            sampler will not run, but read the log file instead, and make
            sure the data and model setting is the same as the previous run.
        """
        if constructor_kwargs is None:
            constructor_kwargs = {}
        else:
            constructor_kwargs = dict(constructor_kwargs)

        if termination_kwargs is None:
            termination_kwargs = {}
        else:
            termination_kwargs = dict(termination_kwargs)

        log_prob, transform, param_names = reparam_loglike(
            self._helper.numpyro_model,
            jax.random.PRNGKey(self._helper.seed['mcmc']),
        )

        @jax.jit
        def log_prob_(params):
            logp = log_prob(dict(zip(param_names, params)))
            if ignore_nan:
                return jnp.nan_to_num(logp, nan=-1e300)
            else:
                return logp

        # if parallel:
        #     ncore = jax.local_device_count()
        #     devices = create_device_mesh((ncore,))
        #     mesh = Mesh(devices, axis_names=('i',))
        #     pi = PartitionSpec('i')
        #     log_prob_parallel = shard_map(
        #         f=jax.jit(jax.vmap(log_prob_)),
        #         mesh=mesh,
        #         in_specs=(pi,),
        #         out_specs=pi,
        #         check_rep=False,
        #     )
        #
        #     def log_prob_(params):
        #         pad = 0
        #         if len(params) % ncore:
        #             pad = ncore - len(params) % ncore
        #             params = np.pad(params, (0, pad), mode='edge')
        #         return log_prob_parallel(params)[: len(params) - pad]
        #
        #     constructor_kwargs['ndraw_min'] = n_batch * ncore
        #     constructor_kwargs['ndraw_max'] = n_batch * ncore
        #     constructor_kwargs['num_test_samples'] = ncore
        #     constructor_kwargs['vectorized'] = True

        @jax.vmap
        @jax.jit
        def transform_(samples):
            base_names = [name + '_base' for name in param_names]
            return transform(dict(zip(base_names, samples)))

        sampler = ultranest.ReactiveNestedSampler(
            param_names=param_names, loglike=log_prob_, **constructor_kwargs
        )
        sampler._transform_back = transform_

        if read_file is None:
            print('Start nested sampling...')
            t0 = time.time()
            sampler.run(min_ess=int(ess), **termination_kwargs)
            print(f'Sampling cost {time.time() - t0:.2f} s')
        else:
            read_file = dict(read_file)
            log_dir = read_file['log_dir']
            verbose = read_file.get('verbose', False)
            x_dim = sampler.x_dim
            sequence, final = ultranest.read_file(
                log_dir, x_dim, verbose=verbose
            )
            sampler.results = sequence | final
        return PosteriorResult(sampler, self._helper, self)

    def nautilus(
        self,
        ess: int = 10000,
        ignore_nan: bool = True,
        parallel: bool = True,
        n_batch: int = 5000,
        *,
        constructor_kwargs: dict | None = None,
        termination_kwargs: dict | None = None,
    ) -> PosteriorResult:
        """Run the Nested Sampler of :mod:`nautilus`.

        Parameters
        ----------
        ess : int, optional
            The desired effective sample size.
        ignore_nan : bool, optional
            Whether to transform a NaN log probability to a large negative
            number (-1e300). The default is True.

            .. warning ::
                Setting ``ignore_nan=True`` may fail to spot potential issues
                with model computation.
        parallel : bool, optional
            Whether to parallelize likelihood evaluation. The default is True.
        n_batch : int, optional
            Number of likelihood evaluations that are performed at each step
            for each core when `parallel` is True. The default is 5000.
        constructor_kwargs : dict, optional
            Extra parameters passed to
            :class:`nautilus.Sampler`.
        termination_kwargs : dict, optional
            Extra parameters passed to
            :class:`nautilus.Sampler.run()`.
        """
        if constructor_kwargs is None:
            constructor_kwargs = {}
        else:
            constructor_kwargs = dict(constructor_kwargs)

        if termination_kwargs is None:
            termination_kwargs = {}
        else:
            termination_kwargs = dict(termination_kwargs)

        log_prob, transform, param_names = reparam_loglike(
            self._helper.numpyro_model,
            jax.random.PRNGKey(self._helper.seed['mcmc']),
        )

        @jax.jit
        def log_prob_(params):
            logp = log_prob(dict(zip(param_names, params)))
            if ignore_nan:
                return jnp.nan_to_num(logp, nan=-1e300)
            else:
                return logp

        if parallel:
            ncore = jax.local_device_count()
            devices = create_device_mesh((ncore,))
            mesh = Mesh(devices, axis_names=('i',))
            pi = PartitionSpec('i')
            log_prob_ = shard_map(
                f=jax.jit(jax.vmap(log_prob_)),
                mesh=mesh,
                in_specs=(pi,),
                out_specs=pi,
                check_rep=False,
            )
            constructor_kwargs['n_batch'] = n_batch * ncore

        @jax.vmap
        @jax.jit
        def transform_(samples):
            base_names = [name + '_base' for name in param_names]
            return transform(dict(zip(base_names, samples)))

        prior = nautilus.Prior()
        for param in param_names:
            prior.add_parameter(param)

        constructor_kwargs['pass_dict'] = False
        constructor_kwargs['seed'] = self._helper.seed['mcmc']
        constructor_kwargs['vectorized'] = bool(parallel)
        constructor_kwargs.setdefault('pool', (None, jax.local_device_count()))
        sampler = nautilus.Sampler(
            prior=prior,
            likelihood=log_prob_,
            **constructor_kwargs,
        )

        termination_kwargs['discard_exploration'] = True
        termination_kwargs.setdefault('verbose', True)
        print('Start nested sampling...')
        t0 = time.time()
        success = sampler.run(n_eff=int(ess), **termination_kwargs)
        if success:
            print(f'Sampling cost {time.time() - t0:.2f} s')
            sampler._transform_back = transform_
            return PosteriorResult(sampler, self._helper, self)
        else:
            raise RuntimeError(
                'Sampling failed due to limits were reached, please set a '
                'larger `n_like_max` or `timeout`. You can also resume the '
                'sampler from previous one, providing `filepath` and `resume`.'
            )

    def aies(
        self,
        warmup=2000,
        samples=20000,
        chains: int | None = None,
        init: dict[str, float] | None = None,
        chain_method: str = 'vectorized',
        progress: bool = True,
        moves: dict | None = None,
        **aies_kwargs: dict,
    ) -> PosteriorResult:
        """Affine-Invariant Ensemble Sampling (AIES) of :mod:`numpyro`.

        Affine-invariant ensemble sampling [1]_ is a gradient-free method
        that informs Metropolis-Hastings proposals by sharing information
        between chains. Suitable for low to moderate dimensional models.
        Generally, num_chains should be at least twice the dimensionality
        of the model.

        .. note::
            This kernel must be used with even `num_chains` > 1 and
            ``chain_method='vectorized'``.

        Parameters
        ----------
        warmup : int, optional
            Number of warmup steps.
        samples : int, optional
            Number of samples to generate from each chain.
        chains : int, optional
            Number of MCMC chains to run. Defaults to 4 * `D`, where `D` is
            the dimension of model parameters.
        init : dict, optional
            Initial parameter for sampler to start from.
        chain_method : str, optional
            The chain method passed to :class:`numpyro.inf.MCMC`.
        progress : bool, optional
            Whether to show progress bar during sampling. The default is True.
        moves : dict, optional
            Moves for the sampler.
        **aies_kwargs : dict
            Extra parameters passed to :class:`numpyro.infer.AIES`.

        Returns
        -------
        PosteriorResult
            The posterior sampling result.

        References
        ----------
        .. [1] *emcee: The MCMC Hammer*
               (https://iopscience.iop.org/article/10.1086/670067),
               Daniel Foreman-Mackey, David W. Hogg, Dustin Lang,
               and Jonathan Goodman.
        """
        if chains is None:
            chains = 4 * len(self._helper.params_names['free'])
        else:
            chains = int(chains)

        # TODO: option to let sampler starting from MLE
        if init is None:
            init = self._helper.free_default['constr_dic']
        else:
            init = self._helper.free_default['constr_dic'] | dict(init)
        init = self._helper.constr_dic_to_unconstr_arr(init)
        rng = np.random.default_rng(self._helper.seed['mcmc'])
        jitter = 0.1 * np.abs(init)
        low = init - jitter
        high = init + jitter
        init = rng.uniform(low, high, size=(chains, len(init)))
        init = dict(zip(self._helper.params_names['free'], init.T))

        aies_kwargs['model'] = self._helper.numpyro_model
        if moves is None:
            aies_kwargs['moves'] = {
                AIES.DEMove(): 0.5,
                AIES.StretchMove(): 0.5,
            }
        else:
            aies_kwargs['moves'] = moves

        sampler = MCMC(
            AIES(**aies_kwargs),
            num_warmup=warmup,
            num_samples=samples,
            num_chains=chains,
            chain_method=chain_method,
            progress_bar=progress,
        )

        sampler.run(
            rng_key=jax.random.PRNGKey(self._helper.seed['mcmc']),
            init_params=init,
        )
        return PosteriorResult(sampler, self._helper, self)



from elisa.models.parameter import UniformParameter


def test_param_name():
    a = UniformParameter('a', 1.5, 0.0, 2.0)
    a2 = UniformParameter('a', 0.5, 0.0, 2.0)
    b = a + a2
    assert str(b) == "a + a'"
    assert b.default == 2.0



import numpy as np

from elisa import BayesFit, MaxLikeFit
from elisa.models import PowerLaw


def test_trivial_max_like_fit():
    # Setup simulation configuration
    seed = 42
    emin = 1.0
    emax = 100.0
    nbins = 200
    photon_egrid = np.linspace(emin, emax, nbins + 1)
    channel_emin = photon_egrid[:-1]
    channel_emax = photon_egrid[1:]
    response_matrix = np.eye(nbins)
    spec_exposure = 50.0

    # Setup model and simulate data
    alpha = 0.0
    K = 10.0  # photon flux between emin and emax, when alpha = 0
    model = PowerLaw(K=[K], alpha=alpha)
    compiled_model = model.compile()
    data = compiled_model.simulate(
        photon_egrid=photon_egrid,
        channel_emin=channel_emin,
        channel_emax=channel_emax,
        response_matrix=response_matrix,
        spec_exposure=spec_exposure,
        spec_poisson=True,
        seed=seed,
    )

    # Get mle result
    result = MaxLikeFit(data, model).mle()

    # Check that the fit result is correct,
    # note that the analytic result is known for alpha = 0 and uniform egrid
    mle_fit, err_fit = result.mle['PowerLaw.K']
    mle_analytic = np.mean(data.ce)
    de = np.diff(photon_egrid)[0]
    err_analytic = np.sqrt(mle_analytic / nbins / spec_exposure / de)
    ci = result.ci().errors['PowerLaw.K']
    assert np.isclose(mle_fit, mle_analytic)
    assert np.isclose(err_fit, err_analytic)
    assert np.isclose(ci[0], -err_analytic, rtol=1e-3, atol=0)
    assert np.isclose(ci[1], err_analytic, rtol=1e-3, atol=0)
    assert result.ndata['total'] == nbins
    assert result.dof == nbins - 1

    # Check various methods of mle result
    result.boot(1000)
    result.ci(method='boot')
    result.flux(1, 2)
    result.lumin(1, 10000, z=1)
    result.eiso(1, 10000, z=1, duration=spec_exposure)
    result.summary()
    result.plot()
    result.plot('data ne ene eene Fv vFv rq pit corner')
    _ = result.deviance
    _ = result.aic
    _ = result.bic
    assert all(i > 0.05 for i in result.gof.values())

    plotter = result.plot
    plotter.plot_qq('rd')
    plotter.plot_qq('rp')
    plotter.plot_qq('rq')


def test_trivial_bayes_fit():
    # Setup simulation configuration
    seed = 42
    emin = 1.0
    emax = 100.0
    nbins = 200
    photon_egrid = np.linspace(emin, emax, nbins + 1)
    channel_emin = photon_egrid[:-1]
    channel_emax = photon_egrid[1:]
    response_matrix = np.eye(nbins)
    spec_exposure = 50.0

    # Setup model and simulate data
    alpha = 0.5
    K = 10.0
    model = PowerLaw(K=[K], alpha=[alpha])
    model.PowerLaw.K.log = True
    compiled_model = model.compile()
    data = compiled_model.simulate(
        photon_egrid=photon_egrid,
        channel_emin=channel_emin,
        channel_emax=channel_emax,
        response_matrix=response_matrix,
        spec_exposure=spec_exposure,
        spec_poisson=True,
        seed=seed,
    )

    # Get Bayesian fit result, i.e. posterior
    result = BayesFit(data, model).nuts()

    # check the true parameters values are within the 90% CI
    ci = result.ci(cl=0.9).intervals
    assert ci['PowerLaw.K'][0] < K < ci['PowerLaw.K'][1]
    assert ci['PowerLaw.alpha'][0] < alpha < ci['PowerLaw.alpha'][1]

    # Check various methods of posterior result
    assert result.ndata['total'] == nbins
    assert result.dof == nbins - 2

    result.ppc(1000)
    result.flux(1, 2)
    result.lumin(1, 10000, z=1)
    result.eiso(1, 10000, z=1, duration=spec_exposure)
    result.summary()
    result.plot()
    result.plot('data ne ene eene Fv vFv rq pit corner khat')
    _ = result.deviance
    _ = result.loo
    _ = result.waic
    assert all(i > 0.05 for i in result.gof.values())

    plotter = result.plot
    plotter.plot_qq('rd')
    plotter.plot_qq('rp')
    plotter.plot_qq('rq')



import numpy as np

from elisa.data.grouping import significance_gv, significance_lima
from elisa.models import PowerLaw


def test_data_plot():
    # Setup simulation configuration
    seed = 42
    emin = 1.0
    emax = 100.0
    nbins = 200
    photon_egrid = np.linspace(emin, emax, nbins + 1)
    channel_emin = photon_egrid[:-1]
    channel_emax = photon_egrid[1:]
    response_matrix = np.eye(nbins)
    spec_exposure = 50.0

    # Setup model and simulate data
    alpha = 0.0
    K = 10.0  # photon flux between emin and emax, when alpha = 0
    model = PowerLaw(K=[K], alpha=alpha)
    compiled_model = model.compile()
    data = compiled_model.simulate(
        photon_egrid=photon_egrid,
        channel_emin=channel_emin,
        channel_emax=channel_emax,
        response_matrix=response_matrix,
        spec_exposure=spec_exposure,
        spec_poisson=True,
        back_counts=np.full(nbins, 5),
        back_exposure=1.0,
        back_poisson=True,
        seed=seed,
    )

    # Check the plot methods
    data.plot_spec()
    data.plot_effective_area()
    data.plot_matrix()


def test_data_grouping():
    # Setup simulation configuration
    seed = 42
    emin = 1.0
    emax = 100.0
    nbins = 200
    photon_egrid = np.linspace(emin, emax, nbins + 1)
    channel_emin = photon_egrid[:-1]
    channel_emax = photon_egrid[1:]
    response_matrix = np.eye(nbins)
    spec_exposure = 50.0

    # Setup model and simulate data
    alpha = 0.0
    K = 10.0  # photon flux between emin and emax, when alpha = 0
    model = PowerLaw(K=[K], alpha=alpha)
    compiled_model = model.compile()
    data = compiled_model.simulate(
        photon_egrid=photon_egrid,
        channel_emin=channel_emin,
        channel_emax=channel_emax,
        response_matrix=response_matrix,
        spec_exposure=spec_exposure,
        spec_poisson=True,
        back_counts=np.full(nbins, 10),
        back_exposure=2.0,
        back_poisson=True,
        seed=seed,
    )

    scale = 6
    data.group('const', scale)
    assert data.channel.size == nbins // scale

    scale = 1
    data.group('min', scale)
    assert np.all(data.spec_counts >= scale)

    scale = 1
    data.group('sig', scale)
    sig = significance_lima(
        data.spec_counts, data.back_counts, data.back_ratio
    )
    assert np.all(sig >= scale)

    scale = 10
    data.group('bmin', scale)
    assert np.all(data.back_counts >= scale)

    scale = 1
    data.group('bsig', scale)
    sig = data.back_counts / data.back_errors
    assert np.all(sig >= scale)

    data.group('opt')
    assert data.channel.size == nbins

    scale = 1
    data.group('optmin', scale)
    assert np.all(data.spec_counts >= scale)

    scale = 10
    data.group('optbmin', scale)
    assert np.all(data.back_counts >= scale)

    scale = 1
    data.group('optsig', scale)
    sig = significance_lima(
        data.spec_counts, data.back_counts, data.back_ratio
    )
    assert np.all(sig >= scale)

    scale = 1
    data.group('optbsig', scale)
    sig = data.back_counts / data.back_errors
    assert np.all(sig >= scale)

    data = compiled_model.simulate(
        photon_egrid=photon_egrid,
        channel_emin=channel_emin,
        channel_emax=channel_emax,
        response_matrix=response_matrix,
        spec_exposure=spec_exposure,
        spec_poisson=True,
        back_counts=np.full(nbins, 10),
        back_errors=np.full(nbins, 2),
        back_exposure=2.0,
        back_poisson=False,
        seed=seed,
    )

    scale = 1
    data.group('sig', scale)
    sig = significance_gv(
        data.spec_counts, data.back_counts, data.back_errors, data.back_ratio
    )
    assert np.all(sig >= scale)

    scale = 1
    data.group('optbsig', scale)
    sig = significance_gv(
        data.spec_counts, data.back_counts, data.back_errors, data.back_ratio
    )
    assert np.all(sig >= scale)



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0
"""Nested sampling of jaxns.

This module is adapted from
https://github.com/pyro-ppl/numpyro/raw/master/numpyro/contrib/nested_sampling.py

"""
from __future__ import annotations

from functools import singledispatch

import jax
import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
import tensorflow_probability.substrates.jax as tfp
from jax import random
from numpyro.handlers import reparam, seed, trace
from numpyro.infer import Predictive
from numpyro.infer.reparam import Reparam
from numpyro.infer.util import _guess_max_plate_nesting, _validate_model, log_density


tfpd = tfp.distributions


@singledispatch
def uniform_reparam_transform(d):
    """
    A helper for :class:`UniformReparam` to get the transform that transforms
    a uniform distribution over a unit hypercube to the target distribution `d`.
    """
    if isinstance(d, dist.TransformedDistribution):
        outer_transform = dist.transforms.ComposeTransform(d.transforms)
        return lambda q: outer_transform(uniform_reparam_transform(d.base_dist)(q))

    if isinstance(
        d, (dist.Independent, dist.ExpandedDistribution, dist.MaskedDistribution)
    ):
        return lambda q: uniform_reparam_transform(d.base_dist)(q)

    return d.icdf


@uniform_reparam_transform.register(dist.MultivariateNormal)
def _(d):
    outer_transform = dist.transforms.LowerCholeskyAffine(d.loc, d.scale_tril)
    return lambda q: outer_transform(dist.Normal(0, 1).icdf(q))


@uniform_reparam_transform.register(dist.BernoulliLogits)
@uniform_reparam_transform.register(dist.BernoulliProbs)
def _(d):
    def transform(q):
        x = q < d.probs
        return x.astype(jnp.result_type(x, int))

    return transform


@uniform_reparam_transform.register(dist.CategoricalLogits)
@uniform_reparam_transform.register(dist.CategoricalProbs)
def _(d):
    return lambda q: jnp.sum(jnp.cumsum(d.probs, axis=-1) < q[..., None], axis=-1)


@uniform_reparam_transform.register(dist.Dirichlet)
def _(d):
    gamma_dist = dist.Gamma(d.concentration)

    def transform_fn(q):
        # NB: icdf is not available yet for Gamma distribution
        # so this will raise an NotImplementedError for now.
        # We will need scipy.special.gammaincinv, which is not available yet in JAX
        # see issue: https://github.com/google/jax/issues/5350
        # TODO: consider wrap jaxns GammaPrior transform implementation
        gammas = uniform_reparam_transform(gamma_dist)(q)
        return gammas / gammas.sum(-1, keepdims=True)

    return transform_fn


class UniformReparam(Reparam):
    """
    Reparameterize a distribution to a Uniform over the unit hypercube.

    Most univariate distribution uses Inverse CDF for the reparameterization.
    """

    def __call__(self, name, fn, obs):
        assert obs is None, "TransformReparam does not support observe statements"
        shape = fn.shape()
        fn, expand_shape, event_dim = self._unwrap(fn)
        transform = uniform_reparam_transform(fn)
        tiny = jnp.finfo(jnp.result_type(float)).tiny

        x = numpyro.sample(
            "{}_base".format(name),
            dist.Uniform(tiny, 1).expand(shape).to_event(event_dim).mask(False),
        )
        # Simulate a numpyro.deterministic() site.
        return None, transform(x)


class NestedSampler:
    """
    (EXPERIMENTAL) A wrapper for `jaxns <https://github.com/Joshuaalbert/jaxns>`_ ,
    a nested sampling package based on JAX.

    See reference [1] for details on the meaning of each parameter.
    Please consider citing this reference if you use the nested sampler in your research.

    .. note:: To enumerate over a discrete latent variable, you can add the keyword
        `infer={"enumerate": "parallel"}` to the corresponding `sample` statement.

    .. note:: To improve the performance, please consider enabling x64 mode at the beginning
        of your NumPyro program ``numpyro.enable_x64()``.

    **References**

    1. *JAXNS: a high-performance nested sampling package based on JAX*,
       Joshua G. Albert (https://arxiv.org/abs/2012.15286)

    :param callable model: a call with NumPyro primitives
    :param dict constructor_kwargs: additional keyword arguments to construct an upstream
        :class:`jaxns.NestedSampler` instance.
    :param dict termination_kwargs: keyword arguments to terminate the sampler. Please
        refer to the upstream :meth:`jaxns.NestedSampler.__call__` method.

    **Example**

    .. doctest::

        >>> from jax import random
        >>> import jax.numpy as jnp
        >>> import numpyro
        >>> import numpyro.distributions as dist
        >>> from numpyro.contrib.nested_sampling import NestedSampler

        >>> true_coefs = jnp.array([1., 2., 3.])
        >>> data = random.normal(random.PRNGKey(0), (2000, 3))
        >>> labels = dist.Bernoulli(logits=(true_coefs * data).sum(-1)).sample(random.PRNGKey(1))
        >>>
        >>> def model(data, labels):
        ...     coefs = numpyro.sample('coefs', dist.Normal(0, 1).expand([3]))
        ...     intercept = numpyro.sample('intercept', dist.Normal(0., 10.))
        ...     return numpyro.sample('y', dist.Bernoulli(logits=(coefs * data + intercept).sum(-1)),
        ...                           obs=labels)
        >>>
        >>> ns = NestedSampler(model)
        >>> ns.run(random.PRNGKey(2), data, labels)
        >>> samples = ns.get_samples(random.PRNGKey(3), num_samples=1000)
        >>> assert jnp.mean(jnp.abs(samples['intercept'])) < 0.05
        >>> print(jnp.mean(samples['coefs'], axis=0))  # doctest: +SKIP
        [0.93661342 1.95034876 2.86123884]
    """

    def __init__(
        self,
        model,
        *,
        constructor_kwargs=None,
        termination_kwargs=None,
    ):
        from jaxns.utils import NestedSamplerResults

        self.model = model
        self.constructor_kwargs = (
            constructor_kwargs if constructor_kwargs is not None else {}
        )
        self.termination_kwargs = (
            termination_kwargs if termination_kwargs is not None else {}
        )
        self._samples = None
        self._log_weights = None
        self._results: NestedSamplerResults | None = None

    def run(self, rng_key, *args, **kwargs):
        """
        Run the nested samplers and collect weighted samples.

        :param random.PRNGKey rng_key: Random number generator key to be used for the sampling.
        :param args: The arguments needed by the `model`.
        :param kwargs: The keyword arguments needed by the `model`.
        """
        from jaxns import DefaultNestedSampler, Model, Prior, TerminationCondition

        rng_sampling, rng_predictive = random.split(rng_key)
        # reparam the model so that latent sites have Uniform(0, 1) priors
        seeded = jax.jit(seed(self.model, rng_key))
        prototype_trace = trace(seeded).get_trace(*args, **kwargs)
        param_names = [
            site["name"]
            for site in prototype_trace.values()
            if site["type"] == "sample"
            and not site["is_observed"]
            and site["infer"].get("enumerate", "") != "parallel"
        ]
        deterministics = [
            site["name"]
            for site in prototype_trace.values()
            if site["type"] == "deterministic"
        ]
        reparam_model = reparam(
            self.model, config={k: UniformReparam() for k in param_names}
        )

        # enable enumerate if needed
        has_enum = any(
            site["type"] == "sample"
            and site["infer"].get("enumerate", "") == "parallel"
            for site in prototype_trace.values()
        )
        if has_enum:
            from numpyro.contrib.funsor import enum, log_density as log_density_

            max_plate_nesting = _guess_max_plate_nesting(prototype_trace)
            _validate_model(prototype_trace)
            reparam_model = enum(reparam_model, -max_plate_nesting - 1)
        else:
            log_density_ = log_density

        # Jaxns requires loglikelihood function to have explicit signatures.
        local_dict = {}
        loglik_fn_def = """def loglik_fn(params):\n
        \tparams = {k + '_base': v for k, v in params.items()}\n
        \treturn log_density_(reparam_model, args, kwargs, params)[0]
        """
        exec(loglik_fn_def, locals(), local_dict)
        loglik_fn = local_dict["loglik_fn"]

        # use NestedSampler with identity prior chain
        def prior_model():
            params = {}
            for name in param_names:
                shape = prototype_trace[name]["fn"].shape()
                param = yield Prior(
                    tfpd.Uniform(low=jnp.zeros(shape), high=jnp.ones(shape)),
                    name=name + "_base",
                )
                params[name] = param
            return params

        model = Model(prior_model=prior_model, log_likelihood=loglik_fn)

        default_constructor_kwargs = dict(
            num_live_points=model.U_ndims * 25,
            num_parallel_workers=1,
            max_samples=1e4,
        )
        default_termination_kwargs = dict(dlogZ=1e-4)
        # Fill-in missing values with defaults. This allows user to inspect what was actually used by inspecting
        # these dictionaries
        list(
            map(
                lambda item: self.constructor_kwargs.setdefault(*item),
                default_constructor_kwargs.items(),
            )
        )
        list(
            map(
                lambda item: self.termination_kwargs.setdefault(*item),
                default_termination_kwargs.items(),
            )
        )

        default_ns = DefaultNestedSampler(
            model=model,
            **self.constructor_kwargs,
        )

        # TODO: check if this is necessary
        # jit when num_parallel_workers is 1
        if self.constructor_kwargs['num_parallel_workers'] == 1:
            run_default_ns = jax.jit(default_ns)
        else:
            run_default_ns = default_ns

        termination_reason, state = run_default_ns(
            rng_sampling, term_cond=TerminationCondition(**self.termination_kwargs)
        )
        results = default_ns.to_results(
            termination_reason=termination_reason, state=state
        )

        # transform base samples back to original domains
        # Here we only transform the first valid num_samples samples
        # NB: the number of weighted samples obtained from jaxns is results.num_samples
        # and only the first num_samples values of results.samples are valid.
        num_samples = results.total_num_samples
        samples = results.samples
        predictive = Predictive(
            reparam_model, samples, return_sites=param_names + deterministics
        )
        samples = predictive(rng_predictive, *args, **kwargs)
        # replace base samples in jaxns results by transformed samples
        self._results = results._replace(samples=samples)

    def get_samples(self, rng_key, num_samples):
        """
        Draws samples from the weighted samples collected from the run.

        :param random.PRNGKey rng_key: Random number generator key to be used to draw samples.
        :param int num_samples: The number of samples.
        :return: a dict of posterior samples
        """
        from jaxns import resample

        if self._results is None:
            raise RuntimeError(
                "NestedSampler.run(...) method should be called first to obtain results."
            )
        weighted_samples, sample_weights = self.get_weighted_samples()
        return resample(
            rng_key, weighted_samples, sample_weights, S=num_samples, replace=True
        )

    def get_weighted_samples(self):
        """
        Gets weighted samples and their corresponding log weights.
        """
        if self._results is None:
            raise RuntimeError(
                "NestedSampler.run(...) method should be called first to obtain results."
            )

        return self._results.samples, self._results.log_dp_mean

    def print_summary(self):
        """
        Print summary of the result. This is a wrapper of :func:`jaxns.utils.summary`.
        """
        from jaxns import summary

        if self._results is None:
            raise RuntimeError(
                "NestedSampler.run(...) method should be called first to obtain results."
            )
        summary(self._results)

    def diagnostics(self):
        """
        Plot diagnostics of the result. This is a wrapper of :func:`jaxns.plotting.plot_diagnostics`
        and :func:`jaxns.plotting.plot_cornerplot`.
        """
        from jaxns import plot_cornerplot, plot_diagnostics

        if self._results is None:
            raise RuntimeError(
                "NestedSampler.run(...) method should be called first to obtain results."
            )
        plot_diagnostics(self._results)
        plot_cornerplot(self._results)


def reparam_loglike(model, rng_key, *args, **kwargs):
    rng_sampling, rng_predictive = random.split(rng_key)
    # reparam the model so that latent sites have Uniform(0, 1) priors
    seeded = jax.jit(seed(model, rng_key))
    prototype_trace = trace(seeded).get_trace(*args, **kwargs)
    param_names = [
        site['name']
        for site in prototype_trace.values()
        if site['type'] == 'sample'
        and not site['is_observed']
        and site['infer'].get('enumerate', '') != 'parallel'
    ]
    deterministics = [
        site['name']
        for site in prototype_trace.values()
        if site['type'] == 'deterministic'
    ]
    reparam_model = reparam(
        model, config={k: UniformReparam() for k in param_names}
    )

    # enable enumerate if needed
    has_enum = any(
        site['type'] == 'sample'
        and site['infer'].get('enumerate', '') == 'parallel'
        for site in prototype_trace.values()
    )
    if has_enum:
        from numpyro.contrib.funsor import enum, log_density as log_density_

        max_plate_nesting = _guess_max_plate_nesting(prototype_trace)
        _validate_model(prototype_trace)
        reparam_model = enum(reparam_model, -max_plate_nesting - 1)
    else:
        log_density_ = log_density

    # Jaxns requires loglikelihood function to have explicit signatures.
    local_dict = {}
    loglik_fn_def = """def loglik_fn(params):\n
    \tparams = {k + '_base': v for k, v in params.items()}\n
    \treturn log_density_(reparam_model, args, kwargs, params)[0]
    """
    exec(loglik_fn_def, locals(), local_dict)
    loglik_fn = local_dict['loglik_fn']

    def transform_back(samples):
        predictive = Predictive(
            reparam_model, samples, return_sites=param_names + deterministics
        )
        return predictive(rng_predictive, *args, **kwargs)

    return local_dict['loglik_fn'], transform_back, param_names


class GlobalOptimizer:
    """
    :param callable model: a call with NumPyro primitives
    :param dict constructor_kwargs: additional keyword arguments to construct an upstream
        :class:`jaxns.experimental.DefaultGlobalOptimisation` instance.
    :param dict termination_kwargs: keyword arguments to terminate the sampler. Please
        refer to the upstream :meth:`jaxns.experimental.DefaultGlobalOptimisation.__call__` method.
    """
    def __init__(
        self,
        model,
        *,
        constructor_kwargs: dict | None = None,
        termination_kwargs: dict | None = None
    ):
        from jaxns.experimental import (
            DefaultGlobalOptimisation, GlobalOptimisationResults
        )

        self.model = model
        self.constructor_kwargs = (
            constructor_kwargs if constructor_kwargs is not None else {}
        )
        self.termination_kwargs = (
            termination_kwargs if termination_kwargs is not None else {}
        )
        self._optimizer: DefaultGlobalOptimisation | None = None
        self._results: GlobalOptimisationResults | None = None

    def run(self, rng_key, *args, **kwargs):
        """
        Run the nested samplers and collect weighted samples.

        :param random.PRNGKey rng_key: Random number generator key to be used for the sampling.
        :param args: The arguments needed by the `model`.
        :param kwargs: The keyword arguments needed by the `model`.
        """
        from jaxns import Model, Prior
        from jaxns.experimental import (
            DefaultGlobalOptimisation, GlobalOptimisationTerminationCondition
        )

        rng_sampling, rng_predictive = random.split(rng_key)
        # reparam the model so that latent sites have Uniform(0, 1) priors
        prototype_trace = trace(seed(self.model, rng_key)).get_trace(*args, **kwargs)
        param_names = [
            site["name"]
            for site in prototype_trace.values()
            if site["type"] == "sample"
               and not site["is_observed"]
               and site["infer"].get("enumerate", "") != "parallel"
        ]
        deterministics = [
            site["name"]
            for site in prototype_trace.values()
            if site["type"] == "deterministic"
        ]
        reparam_model = reparam(self.model, config={k: UniformReparam() for k in param_names})

        # enable enumerate if needed
        has_enum = any(
            site["type"] == "sample"
            and site["infer"].get("enumerate", "") == "parallel"
            for site in prototype_trace.values()
        )
        if has_enum:
            from numpyro.contrib.funsor import enum, log_density as log_density_

            max_plate_nesting = _guess_max_plate_nesting(prototype_trace)
            _validate_model(prototype_trace)
            reparam_model = enum(reparam_model, -max_plate_nesting - 1)
        else:
            log_density_ = log_density

        # Jaxns requires loglikelihood function to have explicit signatures.
        local_dict = {}
        loglik_fn_def = """def loglik_fn({}):\n
        \tparams = {{{}}}\n
        \treturn log_density_(reparam_model, args, kwargs, params)[0]
        """.format(
            ", ".join([f"{name.replace('.', '_')}_base" for name in param_names]),
            ", ".join(
                [f"'{name}_base': {name.replace('.', '_')}_base" for name in param_names]
            ),
        )
        exec(loglik_fn_def, locals(), local_dict)
        loglik_fn = local_dict["loglik_fn"]

        # use NestedSampler with identity prior chain
        def prior_model():
            params = []
            for name in param_names:
                shape = prototype_trace[name]["fn"].shape()
                param = yield Prior(
                    tfpd.Uniform(low=jnp.zeros(shape), high=jnp.ones(shape)),
                    name=name + "_base",
                )
                params.append(param)
            return tuple(params)

        model = Model(prior_model=prior_model, log_likelihood=loglik_fn)

        default_constructor_kwargs = dict(
            num_search_chains=model.U_ndims * 20,
            num_parallel_workers=1,
            s=1,
            k=0,
            gradient_slice=False,
        )
        default_termination_kwargs = {}
        # Fill-in missing values with defaults. This allows user to inspect what was actually used by inspecting
        # these dictionaries
        list(
            map(
                lambda item: self.constructor_kwargs.setdefault(*item),
                default_constructor_kwargs.items(),
            )
        )
        list(
            map(
                lambda item: self.termination_kwargs.setdefault(*item),
                default_termination_kwargs.items(),
            )
        )

        self._optimizer = DefaultGlobalOptimisation(
            model=model,
            **self.constructor_kwargs,
        )

        self._results = self._optimizer(
            rng_sampling,
            term_cond=GlobalOptimisationTerminationCondition(**self.termination_kwargs)
        )

        return self._results

    def print_summary(self, io=None):
        self._optimizer.summary(self._results, io)



import jax
import numpy as np
from astropy.cosmology import Planck18
from astropy.units import Unit

from elisa import ConstantValue, ParamConfig, PyAnaInt, PyNumInt
from elisa.models import PhAbs, PLPhFlux, PowerLaw, ZAShift


def test_name():
    model1 = PowerLaw() + PowerLaw()
    model2 = PhAbs() * PowerLaw()
    model3 = ZAShift(1.0)(PowerLaw())

    assert model1.name == 'PowerLaw + PowerLaw_2'
    assert model2.name == 'PhAbs * PowerLaw'
    assert model3.name == 'ZAShift(PowerLaw)'


def test_str_repr():
    model = ZAShift(1.0)(PhAbs() * PowerLaw())
    compiled = model.compile()

    assert str(model) == str(compiled) == 'ZAShift(PhAbs * PowerLaw)'
    assert repr(model) == repr(compiled)
    assert model._repr_html_() == compiled._repr_html_()


def test_custom_model():
    class PL1(PyAnaInt):
        _config = (
            ParamConfig('alpha', r'\alpha', '', 1.01, -5.0, 5.0),
            ParamConfig('K', 'K', '', 1.0, 1e-10, 1e10),
        )
        type: str = 'add'

        @staticmethod
        def integral(egrid, params):
            alpha = params['alpha']
            K = params['K']
            one_minus_alpha = 1.0 - alpha
            f = egrid**one_minus_alpha / one_minus_alpha
            return K * (f[1:] - f[:-1])

    class PL2(PyNumInt):
        _config = (
            ParamConfig('alpha', r'\alpha', '', 1.01, -5.0, 5.0),
            ParamConfig('K', 'K', '', 1.0, 1e-10, 1e10),
        )
        type: str = 'add'

        @staticmethod
        def continuum(egrid, params):
            return params['K'] * egrid ** -params['alpha']

    egrid = np.geomspace(1, 10, 1000)
    m0 = PowerLaw().compile()
    m1 = PL1().compile()
    m2 = PL2().compile()
    assert np.allclose(m0.eval(egrid), m1.eval(egrid))
    assert np.allclose(m0.eval(egrid), m2.eval(egrid))

    grad_fn0 = jax.grad(lambda params: m0.eval(egrid, params).sum())
    grad_fn1 = jax.grad(lambda params: m1.eval(egrid, params).sum())
    grad_fn2 = jax.grad(lambda params: m2.eval(egrid, params).sum())
    params0 = {f'{m0.name}.alpha': 1.01, f'{m0.name}.K': 1.0}
    params1 = {f'{m1.name}.alpha': 1.01, f'{m1.name}.K': 1.0}
    params2 = {f'{m2.name}.alpha': 1.01, f'{m2.name}.K': 1.0}
    grad0 = grad_fn0(params0)
    grad1 = grad_fn1(params1)
    grad2 = grad_fn2(params2)
    assert np.allclose(grad0[f'{m0.name}.alpha'], grad1[f'{m1.name}.alpha'])
    assert np.allclose(grad0[f'{m0.name}.K'], grad1[f'{m1.name}.K'])
    assert np.allclose(grad0[f'{m0.name}.alpha'], grad2[f'{m2.name}.alpha'])
    assert np.allclose(grad0[f'{m0.name}.K'], grad2[f'{m2.name}.K'])


def test_lumin_and_eiso():
    def powerlaw(alpha, K, egrid):
        egrid = np.array(egrid)
        one_minus_alpha = 1.0 - alpha
        f = K / one_minus_alpha * np.power(egrid, one_minus_alpha)
        return f[1:] - f[:-1]

    flux_unit = Unit('keV cm^-2 s^-1')

    alpha = 2.5
    K = 1e4  # normalization at 1 keV, ph s^-1 cm^-2 keV^-1
    r = 1.0 * Unit('km')
    area = 4.0 * np.pi * r**2
    exposure = 1.0 * Unit('s')
    flux = powerlaw(alpha - 1.0, K, [1.0, 1e4]) * flux_unit
    eiso = flux * area * exposure
    eiso = eiso.to('erg').value
    lumin = eiso / exposure.value

    z = 3.0
    dc = Planck18.comoving_distance(z).to('km')
    factor = (r / dc) ** 2
    factor = factor.value
    K_obs = K * factor
    exposure_obs = exposure * (1 + z)

    # def zpowerlaw(alpha, K, egrid, z):
    #     egrid = np.array(egrid)
    #     return powerlaw(alpha, K, egrid * (1 + z)) / (1 + z)
    #
    # def zpowerlaw_flux_rest(alpha, K, egrid, z):
    #     alpha = alpha - 1.0
    #     egrid = np.array(egrid) / (1 + z)
    #     return zpowerlaw(alpha, K, egrid, z)
    #
    # area_1 = 4.0 * np.pi * dc**2
    # flux_obs = zpowerlaw_flux_rest(alpha, K_, [1, 10000], z) * flux_unit
    # eiso_obs = flux_obs * area_1 * exposure_obs
    # lumin_obs = eiso_obs / (exposure_obs / (1 + z))
    # eiso_obs = eiso_obs.to('erg').value
    # lumin_obs = lumin_obs.to('erg s^-1').value

    K_param = ConstantValue('K', K_obs)
    model = ZAShift(z)(PowerLaw(K=K_param, alpha=alpha)).compile()
    lumin_calc = model.lumin(1, 10000, z).value
    eiso_calc = model.eiso(1, 10000, z, exposure_obs.value).value

    assert np.allclose(lumin, lumin_calc)
    assert np.allclose(eiso, eiso_calc)


def test_simulation():
    def powerlaw(alpha, K, egrid):
        egrid = np.array(egrid)
        one_minus_alpha = 1.0 - alpha
        f = K / one_minus_alpha * np.power(egrid, one_minus_alpha)
        return f[1:] - f[:-1]

    seed = 42
    emin = 1.0
    emax = 100.0
    nbins = 10
    photon_egrid = np.linspace(emin, emax, nbins + 1)
    channel_emin = photon_egrid[:-1]
    channel_emax = photon_egrid[1:]
    response_matrix = np.eye(nbins)
    spec_exposure = 100.0
    back_exposure = 50.0
    background = np.ones(nbins) * 100.0

    rng = np.random.default_rng(seed)
    back_counts = rng.poisson(background)

    alpha = 2.5
    F = 1000.0  # photon flux between emin and emax
    unscaled_flux = powerlaw(alpha, 1.0, photon_egrid)
    factor = F / powerlaw(alpha, 1.0, [emin, emax])
    source = unscaled_flux * factor * spec_exposure
    total = source + spec_exposure * background / back_exposure
    total_counts = rng.poisson(total)

    model = PLPhFlux(emin, emax, F=F, alpha=alpha).compile()
    data = model.simulate(
        photon_egrid=photon_egrid,
        channel_emin=channel_emin,
        channel_emax=channel_emax,
        response_matrix=response_matrix,
        spec_exposure=spec_exposure,
        spec_poisson=True,
        back_counts=background,
        back_exposure=back_exposure,
        back_poisson=True,
        seed=seed,
    )

    assert np.all(data.spec_counts == total_counts)
    assert np.all(data.back_counts == back_counts)



"""Helper for using scipy.integrate.nquad in JAX.

Contributed by @xiesl97 (https://github.com/xiesl97).
"""

import jax
import jax.numpy as jnp
import numpy as np
from scipy import LowLevelCallable, integrate

try:
    import numba as nb
except ImportError as e:
    raise ImportError(
        'To use this module, please install `numba` package. It can be'
        ' installed with `pip install numba`'
    ) from e


class NQuadTransform:
    def __init__(self, fun):
        self._c_sig = nb.types.double(
            nb.types.intc, nb.types.CPointer(nb.types.double)
        )
        self._fun = fun

    def _cfun(self):
        fun = self._fun

        @nb.cfunc(self._c_sig)
        def cfun(n, _params):
            params = nb.carray(_params, (n,))
            return fun(params)

        return cfun.ctypes

    @staticmethod
    def _nquad(cfun, opts=None, vectorized=False):
        @jax.jit
        def _nquad_scipy(ranges, args):
            ranges = jnp.asarray(ranges)
            args = jnp.asarray(args)

            def _pcb(ranges, args):
                fun_sci_call = LowLevelCallable(cfun)
                result, abserr = integrate.nquad(
                    fun_sci_call, ranges, args, opts, full_output=False
                )
                return jnp.asarray([result, abserr])

            result_shape_dtype = jax.ShapeDtypeStruct(
                shape=(2,), dtype=ranges.dtype
            )
            return jax.pure_callback(
                _pcb, result_shape_dtype, ranges, args, vectorized=vectorized
            )

        return _nquad_scipy

    def _args_grad(self, h=1e-7):
        cfun = self._cfun()
        func_nquad = jax.jit(self._nquad(cfun))

        @jax.custom_jvp
        @jax.jit
        def _fn(ranges, args):
            return func_nquad(ranges, args)[0]

        @_fn.defjvp
        @jax.jit
        def _fn_jvp(primals, tangents):
            ranges, args = primals
            ranges_dot, args_dots = tangents
            primal_out = _fn(ranges, args)
            args_h = args + jnp.eye(len(args)) * h
            primal_dx = jax.vmap(_fn, in_axes=(None, 0))(ranges, args_h)
            primal_grad_dx = (primal_dx - primal_out) / h
            tangent_out = jnp.sum(primal_grad_dx * args_dots)
            return primal_out, tangent_out

        return _fn


if __name__ == '__main__':
    """example 1"""

    @nb.njit
    def f(params):
        x, y, z, d = params
        return np.exp(-(x**2)) + y + z * d

    ranges = jnp.asarray([[0.0, 1.0], [0.0, 1.0]], dtype=jnp.float64)
    args = jnp.asarray([3.0, 4.0], dtype=jnp.float64)
    nqt = NQuadTransform(f)
    cfun = nqt._cfun()
    func_nquad = jax.jit(nqt._nquad(cfun, opts=None, vectorized=False))
    func_arg_grad = nqt._args_grad()
    print(func_nquad(ranges, args))
    print(jax.grad(func_arg_grad)(ranges, args))

    """example 2"""
    from elisa.models.model import AnaIntAdditive, ParamConfig
    from elisa.util.typing import JAXArray, NameValMapping

    @nb.njit
    def bbodyrad(params):
        e, kT, K = params
        return 1.0344e-3 * K * e * e / np.expm1(e / kT)

    nqt = NQuadTransform(bbodyrad)
    func_arg_grad = nqt._args_grad()

    class BlackbodyRad_test(AnaIntAdditive):
        _config = (
            ParamConfig('kT', 'kT', 'keV', 3.0, 1e-4, 200.0),
            ParamConfig('K', 'K', '', 1.0, 1e-10, 1e10),
        )

        @staticmethod
        def integral(egrid: JAXArray, params: NameValMapping) -> JAXArray:
            kT = params['kT']
            K = params['K']

            ranges = jnp.asarray([egrid[:-1], egrid[1:]], dtype=jnp.float64).T
            ranges = jnp.reshape(ranges, (len(ranges), 1, 2))
            args = jnp.asarray([kT, K], dtype=jnp.float64)

            return jax.vmap(func_arg_grad, in_axes=(0, None))(ranges, args)



"""Subsequent analysis of maximum likelihood or Bayesian fit."""

from __future__ import annotations

from abc import ABC, abstractmethod
from importlib import metadata
from typing import TYPE_CHECKING, NamedTuple

import arviz as az
import astropy.units as u
import jax
import jax.numpy as jnp
import nautilus
import numpy as np
import numpyro
import scipy.stats as stats
import ultranest
from astropy.cosmology import Planck18
from iminuit import Minuit
from jax.experimental.mesh_utils import create_device_mesh
from jax.experimental.shard_map import shard_map
from jax.sharding import Mesh, PartitionSpec
from numpyro.infer import MCMC

from elisa.__about__ import __version__
from elisa.infer.helper import check_params
from elisa.infer.nested_sampling import NestedSampler
from elisa.plot.plotter import MLEResultPlotter, PosteriorResultPlotter
from elisa.util.misc import make_pretty_table

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Sequence
    from typing import Literal

    from arviz.stats.stats_utils import ELPDData
    from astropy.cosmology.flrw.lambdacdm import LambdaCDM
    from astropy.units import Quantity as Q
    from iminuit.util import FMin

    from elisa.infer.fit import BayesFit
    from elisa.infer.helper import Helper
    from elisa.plot.plotter import Plotter
    from elisa.util.typing import JAXArray

ReactiveNestedSampler = ultranest.ReactiveNestedSampler
Sampler = nautilus.Sampler


class FitResult(ABC):
    """Fit result."""

    _helper: Helper
    _plotter: Plotter | None
    _flux_fn: Callable
    _lumin_fn: Callable
    _eiso_fn: Callable

    def __init__(self, helper: Helper):
        self._helper = helper

        models = helper.model
        ne = {name: model.ne for name, model in models.items()}
        ene = {name: model.ene for name, model in models.items()}

        def _flux(
            egrid: JAXArray,
            params: dict[str, JAXArray],
            energy: bool,
            comps: bool,
        ) -> dict[str, JAXArray] | dict[str, dict[str, JAXArray]]:
            """Calculate flux."""
            if energy:
                fns = ene
            else:
                fns = ne
            de = jnp.diff(egrid)
            flux = {}
            for name, fn in fns.items():
                f = fn(egrid, params, comps)
                if comps:
                    flux[name] = jax.tree_map(
                        lambda v: jnp.sum(v * de, axis=-1), f
                    )
                else:
                    flux[name] = jnp.sum(f * de, axis=-1)
            return flux

        self._flux_fn = jax.jit(_flux, static_argnums=(2, 3))

    @abstractmethod
    def __repr__(self):
        pass

    @abstractmethod
    def _repr_html_(self):
        pass

    @property
    @abstractmethod
    def plot(self) -> Plotter:
        """Result plotter."""
        pass

    def summary(self, file=None) -> None:
        """Print the summary of fit result.

        Parameters
        ----------
        file: file-like
            An object with a ``write(string)`` method. This is passed to
            :py:func:`print`.
        """
        print(repr(self), file=file)

    @abstractmethod
    def flux(
        self, *args, **kwargs
    ) -> dict[str, jax.Array] | dict[str, dict[str, jax.Array]]:
        pass

    @abstractmethod
    def lumin(
        self, *args, **kwargs
    ) -> dict[str, jax.Array] | dict[str, dict[str, jax.Array]]:
        pass

    @abstractmethod
    def eiso(
        self, *args, **kwargs
    ) -> dict[str, jax.Array] | dict[str, dict[str, jax.Array]]:
        pass

    @property
    def ndata(self) -> dict[str, int]:
        """Data points number."""
        return self._helper.ndata

    @property
    def dof(self) -> int:
        """Degree of freedom."""
        return self._helper.dof

    @property
    @abstractmethod
    def gof(self) -> dict[str, float]:
        """Goodness of fit p-value."""
        pass

    @property
    @abstractmethod
    def _params_dist(self) -> dict[str, JAXArray]:
        pass


class MLEResult(FitResult):
    """Result of maximum likelihood fit."""

    _plotter: MLEResultPlotter | None = None

    def __init__(self, minuit: Minuit, helper: Helper):
        super().__init__(helper)

        self._minuit = minuit

        self._mle_unconstr = jnp.array(minuit.values, float)
        mle, cov = helper.get_mle(self._mle_unconstr)

        if np.isnan(cov).any() and minuit.covariance is not None:
            cov_unconstr = jnp.array(minuit.covariance, float)
            cov = helper.params_covar(self._mle_unconstr, cov_unconstr)

        err = jnp.sqrt(jnp.diagonal(cov))

        # MLE of model params in constrained space
        self._mle = dict(zip(helper.params_names['all'], zip(mle, err)))

        # model deviance at MLE
        self._deviance = jax.jit(helper.deviance)(self._mle_unconstr)

        # model values at MLE
        sites = jax.jit(helper.get_sites)(self._mle_unconstr)
        self._model_values = sites['models']

        # model comparison statistics
        k = self._helper.nparam
        n = self._helper.ndata['total']
        stat = self._deviance['total']
        self._aic = float(stat + k * 2 * (1 + (k + 1) / (n - k - 1)))
        self._bic = float(stat + k * np.log(n))

        # parametric bootstrap result
        self._boot: BootstrapResult | None = None

    def __repr__(self):
        tabs = self._tabs()
        return (
            f'Parameters\n{tabs["params"]}\n\n'
            f'Fit Statistics\n{tabs["stat"]}\n\n'
            f'Information Criterion\n{tabs["ic"]}\n\n'
            f'Fit Status\n{self.status}\n'
        )

    def _repr_html_(self):
        """The repr in Jupyter notebook environment."""
        tabs = self._tabs()
        params_tab = tabs['params'].get_html_string(format=True)
        stat_tab = tabs['stat'].get_html_string(format=True)
        ic_tab = tabs['ic'].get_html_string(format=True)
        status_tab = self.status._repr_html_()
        return (
            '<details open><summary><b>MLE Result</b></summary>'
            '<details open style="padding-left: 1em">'
            f'<summary><b>Parameters</b></summary>{params_tab}</details>'
            '<details open style="padding-left: 1em">'
            f'<summary><b>Fit Statistics</b></summary>{stat_tab}</details>'
            f'<details open style="padding-left: 1em">'
            '<summary><b>Information Criterion</b></summary>'
            f'{ic_tab}</details>'
            '<details style="padding-left: 1em">'
            f'<summary><b>Fit Status</b></summary>{status_tab}</details>'
            '</details>'
        )

    def _tabs(self):
        params_tab = make_pretty_table(
            ['Parameter', 'MLE', 'Error'],
            [(k, f'{v[0]:.4g}', f'{v[1]:.4g}') for k, v in self.mle.items()],
        )
        stat_type = self._helper.statistic
        deviance = self.deviance
        ndata = self.ndata
        rows = [
            [i, f'{stat_type[i]}', f'{deviance[i]:.2f}', ndata[i]]
            for i in self.ndata.keys()
            if i != 'total'
        ]
        rows.append(
            [
                'Total',
                'stat/dof',
                f'{deviance["total"]:.2f}/{self.dof}',
                ndata['total'],
            ]
        )
        names = ['Data', 'Statistic', 'Value', 'Channels']
        stat_tab = make_pretty_table(names, rows)

        rows = [['AIC', f'{self.aic:.2f}'], ['BIC', f'{self.bic:.2f}']]
        names = ['Method', 'Value']
        ic_tab = make_pretty_table(names, rows)
        return {'params': params_tab, 'stat': stat_tab, 'ic': ic_tab}

    @property
    def plot(self) -> MLEResultPlotter:
        if self._plotter is None:
            self._plotter = MLEResultPlotter(self)
        return self._plotter

    def ci(
        self,
        params: str | Iterable[str] | None = None,
        cl: float | int = 1,
        method: Literal['profile', 'boot'] = 'profile',
    ) -> ConfidenceInterval:
        """Calculate confidence intervals.

        Parameters
        ----------
        params : str or sequence of str, optional
            Parameters to calculate confidence intervals. If not specified,
            calculate for parameters of interest.
        cl : float or int, optional
            Confidence level for the confidence interval. If 0 < `cl` < 1, the
            value is interpreted as the confidence level. If `cl` >= 1, it is
            interpreted as the number of standard deviations. For example,
            ``cl=1`` produces a 1-sigma or 68.3% confidence interval.
            The default is 1.
        method : {'profile', 'boot'}, optional
            Method used to calculate confidence. Available options are:

                * ``'profile'``: use Minos algorithm of Minuit to find the
                  confidence intervals based on the profile likelihood
                * ``'boot'``: use parametric bootstrap method to calculate
                  the confidence intervals. :meth:`MLEResult.boot` must be
                  called before using this method.

            The default is ``'profile'``.

        Returns
        -------
        ConfidenceInterval
            The confidence intervals.
        """
        if not self._minuit.valid:
            msg = 'the fit must be valid to calculate confidence interval'
            raise RuntimeError(msg)

        if cl <= 0.0:
            raise ValueError('cl must be non-negative')

        if method not in {'profile', 'boot'}:
            raise ValueError(f'unsupported method: {method}')

        params_names = self._helper.params_names

        params = check_params(params, self._helper)

        params_set = set(params)
        free = params_set.intersection(params_names['free'])
        composite = params_set.intersection(params_names['deterministic'])
        assert free | composite == params_set

        if method == 'profile':
            empty = ({}, {})
            res1 = self._ci_free(free, cl) if free else empty
            res2 = self._ci_composite(composite, cl) if composite else empty
            intervals, status = (r1 | r2 for r1, r2 in zip(res1, res2))

        elif method == 'boot':
            if self._boot is None:
                raise RuntimeError(
                    'before using the bootstrap method to calculate confidence'
                    ' intervals, MLEResult.boot(...) must be called'
                )
            intervals, status = self._ci_boot(params, cl)

        else:
            raise ValueError("method must be either 'profile' or 'boot'")

        mle = {k: v[0] for k, v in self.mle.items()}
        intervals = _format_result(intervals, params)
        errors = {
            k: (intervals[k][0] - mle[k], intervals[k][1] - mle[k])
            for k in params
        }

        return ConfidenceInterval(
            mle=mle,
            intervals=intervals,
            errors=errors,
            cl=1.0 - 2.0 * stats.norm.sf(cl) if cl >= 1.0 else cl,
            method=method,
            status=status,
        )

    def _calc_flux(
        self,
        egrid: JAXArray,
        cl: float | int,
        energy: bool,
        comps: bool,
        params: dict[str, JAXArray] | None,
    ) -> dict[str, Q | float]:
        """Calculate flux."""
        boot_params = self._params_dist

        if boot_params is None:
            raise RuntimeError(
                'MLEResult.boot(...) must be called before calculating flux'
            )

        if energy:
            unit = u.Unit('erg cm^-2 s^-1')
        else:
            unit = u.Unit('ph cm^-2 s^-1')

        mle_params = {k: v[0] for k, v in self._mle.items()}
        if params is not None:
            mle_params |= params
        mle_flux = self._flux_fn(egrid, mle_params, energy, comps)

        n = [i.size for i in boot_params.values()][0]
        if params is not None:
            params = dict(params)
            params = {k: jnp.full(n, v) for k, v in params.items()}
            boot_params = boot_params | params
        devices = create_device_mesh((jax.local_device_count(),))
        mesh = Mesh(devices, axis_names=('i',))
        p = PartitionSpec()
        pi = PartitionSpec('i')
        fn = shard_map(
            f=self._flux_fn,
            mesh=mesh,
            in_specs=(p, pi, p, p),
            out_specs=pi,
            check_rep=False,
        )
        boot_flux = jax.device_get(fn(egrid, boot_params, energy, comps))

        cl_ = 1.0 - 2.0 * stats.norm.sf(cl) if cl >= 1.0 else cl
        q = 0.5 + np.array([-0.5, 0.5]) * cl_
        ci_fn = lambda x: np.quantile(x, q)
        intervals = jax.tree_map(ci_fn, boot_flux)
        errors = jax.tree_map(
            lambda x, y: (y[0] - x, y[1] - x), mle_flux, intervals
        )
        add_unit = lambda x: x * unit
        return {
            'mle': jax.tree_map(add_unit, mle_flux),
            'intervals': jax.tree_map(add_unit, intervals),
            'errors': jax.tree_map(add_unit, errors),
            'cl': cl_,
            'dist': jax.tree_map(add_unit, boot_flux),
            'n': n,
        }

    def flux(
        self,
        emin: float | int,
        emax: float | int,
        cl: float | int = 1,
        energy: bool = True,
        ngrid: int = 1000,
        comps: bool = False,
        log: bool = True,
        params: dict[str, float | int] | None = None,
    ) -> MLEFlux:
        r"""Calculate the flux of model.

        .. warning::
            The flux is calculated by trapezoidal rule, and is accurate only
            if enough numbers of energy grids are used.

        Parameters
        ----------
        emin : float or int
            Minimum value of energy range, in units of keV.
        emax : float or int
            Maximum value of energy range, in units of keV.
        cl : float or int, optional
            Confidence level for the confidence interval. If 0 < `cl` < 1, the
            value is interpreted as the confidence level. If `cl` >= 1, it is
            interpreted as the number of standard deviations. For example,
            ``cl=1`` produces a 1-sigma or 68.3% confidence interval.
            The default is 1.
        energy : bool, optional
            When True, calculate energy flux in units of erg cm⁻² s⁻¹;
            otherwise calculate photon flux in units of ph cm⁻² s⁻¹.
            The default is True.
        ngrid : int, optional
            The energy grid number to use in integration. The default is 1000.

        Other Parameters
        ----------------
        comps : bool, optional
            Whether to return the result of each component. The default is
            False.
        log : bool, optional
            Whether to use logarithmically regular energy grid. The default is
            True.
        params : dict, optional
            Parameters dict to overwrite the fitted parameters.

        Returns
        -------
        MLEFlux
            The flux of the model.
        """
        if self._boot is None:
            raise RuntimeError(
                'MLEResult.boot(...) must be called before calculating flux'
            )

        if log:
            egrid = jnp.geomspace(emin, emax, ngrid)
        else:
            egrid = jnp.linspace(emin, emax, ngrid)

        flux = self._calc_flux(egrid, cl, energy, comps, params)

        return MLEFlux(
            mle=flux['mle'],
            intervals=flux['intervals'],
            errors=flux['errors'],
            cl=flux['cl'],
            dist=flux['dist'],
            energy=bool(energy),
            n=flux['n'],
        )

    def lumin(
        self,
        emin_rest: float | int,
        emax_rest: float | int,
        z: float | int,
        cl: float | int = 1,
        ngrid: int = 1000,
        comps: bool = False,
        log: bool = True,
        params: dict[str, float | int] | None = None,
        cosmo: LambdaCDM = Planck18,
    ) -> MLELumin:
        """Calculate the luminosity of model.

        .. warning::
            The luminosity is calculated by trapezoidal rule, and is accurate
            only if enough numbers of energy grids are used.

        Parameters
        ----------
        emin_rest : float or int
            Minimum value of rest-frame energy range, in units of keV.
        emax_rest : float or int
            Maximum value of rest-frame energy range, in units of keV.
        z : float or int
            Redshift of the source.
        cl : float or int, optional
            Confidence level for the confidence interval. If 0 < `cl` < 1, the
            value is interpreted as the confidence level. If `cl` >= 1, it is
            interpreted as the number of standard deviations. For example,
            ``cl=1`` produces a 1-sigma or 68.3% confidence interval.
            The default is 1.
        ngrid : int, optional
            The energy grid number to use in integration. The default is 1000.

        Other Parameters
        ----------------
        comps : bool, optional
            Whether to return the result of each component. The default is
            False.
        log : bool, optional
            Whether to use logarithmically regular energy grid. The default is
            True.
        params : dict, optional
            Parameters dict to overwrite the fitted parameters.
        cosmo : LambdaCDM, optional
            Cosmology model used to calculate luminosity. The default is
            Planck18.

        Returns
        -------
        MLELumin
            The luminosity of the model.
        """
        if log:
            egrid = jnp.geomspace(emin_rest, emax_rest, ngrid) / (1.0 + z)
        else:
            egrid = jnp.linspace(emin_rest, emax_rest, ngrid) / (1.0 + z)

        flux = self._calc_flux(egrid, cl, True, comps, params)

        factor = 4.0 * np.pi * cosmo.luminosity_distance(z) ** 2
        to_lumin = lambda x: (x * factor).to('erg s^-1')

        return MLELumin(
            mle=jax.tree_map(to_lumin, flux['mle']),
            intervals=jax.tree_map(to_lumin, flux['intervals']),
            errors=jax.tree_map(to_lumin, flux['errors']),
            cl=flux['cl'],
            dist=jax.tree_map(to_lumin, flux['dist']),
            n=flux['n'],
            z=float(z),
            cosmo=cosmo,
        )

    def eiso(
        self,
        emin_rest: float | int,
        emax_rest: float | int,
        z: float | int,
        duration: float | int,
        cl: float | int = 1,
        ngrid: int = 1000,
        comps: bool = False,
        log: bool = True,
        params: dict[str, float | int] | None = None,
        cosmo: LambdaCDM = Planck18,
    ) -> MLEEIso:
        r"""Calculate the isotropic emission energy of model.

        .. warning::
            The :math:`E_\mathrm{iso}` is calculated by trapezoidal rule,
            and is accurate only if enough numbers of energy grids are used.

        Parameters
        ----------
        emin_rest : float or int
            Minimum value of rest-frame energy range, in units of keV.
        emax_rest : float or int
            Maximum value of rest-frame energy range, in units of keV.
        z : float or int
            Redshift of the source.
        duration : float or int
            Observed duration of the source, in units of seconds.
        cl : float or int, optional
            Confidence level for the confidence interval. If 0 < `cl` < 1,
            the value is interpreted as the confidence level. If `cl` >= 1,
            it is interpreted as the number of standard deviations.
            For example, ``cl=1`` produces a 1-sigma or 68.3% confidence
            interval. The default is 1.
        ngrid : int, optional
            The energy grid number to use in integration. The default is
            1000.

        Other Parameters
        ----------------
        comps : bool, optional
            Whether to return the result of each component. The default is
            False.
        log : bool, optional
            Whether to use logarithmically regular energy grid. The default
            is True.
        params : dict, optional
            Parameters dict to overwrite the fitted parameters.
        cosmo : LambdaCDM, optional
            Cosmology model used to calculate luminosity. The default is
            Planck18.

        Returns
        -------
        MLEEIso
            The isotropic emission energy of the model.
        """
        lumin = self.lumin(
            emin_rest, emax_rest, z, cl, ngrid, comps, log, params, cosmo
        )

        # This includes correction for energy redshift and time dilation.
        factor = duration / (1 + z) * u.s
        to_eiso = lambda x: (x * factor).to('erg')

        return MLEEIso(
            mle=jax.tree_map(to_eiso, lumin.mle),
            intervals=jax.tree_map(to_eiso, lumin.intervals),
            errors=jax.tree_map(to_eiso, lumin.errors),
            cl=lumin.cl,
            dist=jax.tree_map(to_eiso, lumin.dist),
            n=lumin.n,
            z=lumin.z,
            duration=float(duration),
            cosmo=lumin.cosmo,
        )

    def boot(
        self,
        n: int = 10000,
        seed: int | None = None,
        parallel: bool = True,
        progress: bool = True,
        update_rate: int = 50,
    ):
        """Preform parametric bootstrap.

        Parameters
        ----------
        n : int, optional
            Number of parametric bootstraps based on the MLE. The default is
            10000.
        seed : int, optional
            The seed of random number generator used in parametric bootstrap.
        parallel : bool, optional
            Whether to run simulation fit in parallel. The default is True.
        progress : bool, optional
            Whether to display progress bar. The default is True.
        update_rate : int, optional
            The update rate of progress bar. The default is 50.
        """
        # reuse the previous result if all setup is the same
        if self._boot and self._boot.n == n and self._boot.seed == seed:
            return

        helper = self._helper
        n = int(n)
        seed = helper.seed['pred'] if seed is None else int(seed)
        params = {i: self._mle[i][0] for i in helper.params_names['free']}
        models = self._model_values

        # perform parametric bootstrap
        result = helper.simulate_and_fit(
            seed,
            params,
            models,
            n,
            parallel,
            progress,
            update_rate,
            'Bootstrap',
        )
        valid = result.pop('valid')
        result = jax.tree_map(lambda x: x[valid], result)

        self._boot = BootstrapResult(
            mle={k: v[0] for k, v in self.mle.items()},
            data=result['data'],
            models=result['models'],
            params=result['params'],
            deviance=result['deviance'],
            p_value=jax.tree_map(
                lambda obs, sim: np.sum(sim >= obs, axis=0) / len(sim),
                self._deviance,
                result['deviance'],
            ),
            n=n,
            n_valid=np.sum(valid),
            seed=seed,
        )

    @property
    def gof(self) -> dict[str, float]:
        if self._boot is None:
            raise RuntimeError('MLEResult.boot() must be called to assess gof')
        p_value = self._boot.p_value
        p_value = p_value['group'] | {'total': p_value['total']}
        return {k: float(p_value[k]) for k in self.ndata.keys()}

    @property
    def _params_dist(self) -> dict[str, jax.Array] | None:
        """Bootstrapped parameter distribution."""
        boot = self._boot
        if boot is None:
            return None
        n = boot.n_valid - boot.n_valid % jax.local_device_count()
        return {k: v[:n] for k, v in boot.params.items()}

    def _ci_free(self, names: Iterable[str], cl: float | int):
        """Confidence interval of free parameters."""
        self._minuit.minos(*names, cl=cl)
        mle_unconstr = self._minuit.values.to_dict()
        ci_unconstr = self._minuit.merrors

        # values of uninterested free parameters, in unconstrained space
        others = {k: v for k, v in mle_unconstr.items() if k not in names}

        # lower bound
        lower = self._helper.unconstr_dic_to_params_dic(
            {k: mle_unconstr[k] + ci_unconstr[k].lower for k in names} | others
        )

        # upper bound
        upper = self._helper.unconstr_dic_to_params_dic(
            {k: mle_unconstr[k] + ci_unconstr[k].upper for k in names} | others
        )

        interval = {k: (lower[k], upper[k]) for k in names}
        status = {
            k: {
                'valid': (v.lower_valid, v.upper_valid),
                'at_limit': (v.at_lower_limit, v.at_upper_limit),
                'at_max_fcn': (v.at_lower_max_fcn, v.at_upper_max_fcn),
                'new_min': (v.lower_new_min, v.upper_new_min),
            }
            for k, v in ci_unconstr.items()
            if k in names
        }
        return interval, status

    def _ci_composite(self, names: Iterable[str], cl: float | int):
        """Confidence intervals of function of free parameters.

        References
        ----------
        .. [1] Eq.24 of https://doi.org/10.1007/s11222-021-10012-y
        .. [2] https://github.com/vemomoto/vemomoto/blob/master/ci_rvm/ci_rvm/ci_rvm.py#L1455
        """

        def loss_factory(name, mle):
            """Factory to create loss function for composite parameter."""
            rtol = 1e-3
            atol = mle * rtol
            atol_inv = 1.0 / atol

            @jax.jit
            def _(x: np.ndarray):
                assert len(x) - 1 == len(free_params)
                unconstr_dic = dict(zip(free_params, x[1:]))
                value = helper.unconstr_dic_to_params_dic(unconstr_dic)[name]
                s = (value - x[0]) * atol_inv
                return helper.deviance_total(x[1:]) + s * s

            return _

        helper = self._helper
        free_params = helper.params_names['free']

        interval = {}
        status = {}
        for i in names:
            mle_i = self._mle[i][0]
            loss = loss_factory(i, mle_i)
            init = np.array([mle_i, *self._minuit.values], float)
            grad = jax.jit(jax.grad(loss))
            minuit = Minuit(loss, init, grad=grad)
            minuit.strategy = 1
            minuit.migrad()
            minuit.minos(0, cl=cl)
            ci = minuit.merrors[0]
            interval[i] = (mle_i + ci.lower, mle_i + ci.upper)
            status[i] = {
                'valid': (ci.lower_valid, ci.upper_valid),
                'at_limit': (ci.at_lower_limit, ci.at_upper_limit),
                'at_max_fcn': (ci.at_lower_max_fcn, ci.at_upper_max_fcn),
                'new_min': (ci.lower_new_min, ci.upper_new_min),
            }

        return interval, status

    def _ci_boot(self, names: Iterable[str], cl: float | int):
        """Bootstrap confidence interval."""
        cl = 1.0 - 2.0 * stats.norm.sf(cl) if cl >= 1.0 else cl
        boot = self._boot
        interval = {
            k: np.quantile(v, q=(0.5 - 0.5 * cl, 0.5 + 0.5 * cl)).tolist()
            for k, v in boot.params.items()
            if k in names
        }
        status = {'nboot': boot.n_valid, 'seed': boot.seed}
        return interval, status

    @property
    def mle(self) -> dict[str, tuple[float, float]]:
        """MLE and error of parameters."""
        return _format_result(self._mle, self._helper.params_names['all'])

    @property
    def deviance(self) -> dict[str, float]:
        """Deviance of the model at MLE."""
        stat = self._deviance['group'] | {'total': self._deviance['total']}
        stat = {i: float(stat[i]) for i in (*self._helper.data_names, 'total')}
        return stat

    @property
    def aic(self) -> float:
        """Akaike information criterion with sample size correction."""
        return self._aic

    @property
    def bic(self) -> float:
        """Bayesian information criterion."""
        return self._bic

    @property
    def status(self) -> FMin:
        """Fit status of Minuit."""
        return self._minuit.fmin


class PosteriorResult(FitResult):
    """Result obtained from Bayesian fit."""

    _plotter: PosteriorResultPlotter | None = None
    _idata: az.InferenceData
    _deviance: dict | None = None
    _mle_result: dict | None = None
    _ppc: PPCResult | None = None
    _loo: az.stats.stats_utils.ELPDData | None = None
    _waic: az.stats.stats_utils.ELPDData | None = None
    _rhat: dict[str, float] | None = None
    _divergence: int | None = None
    _pit: dict[str, tuple] | None = None
    _params: dict[str, JAXArray] | None = None
    _info_tabs: dict | None = None

    def __init__(
        self,
        sampler: MCMC | NestedSampler | ReactiveNestedSampler | Sampler,
        helper: Helper,
        fit: BayesFit,
    ):
        if not isinstance(
            sampler, (MCMC, NestedSampler, ReactiveNestedSampler, Sampler)
        ):
            raise ValueError(f'unknown sampler type {type(sampler)}')

        super().__init__(helper)
        self._fit = fit
        if isinstance(sampler, MCMC):
            self._init_from_numpyro(sampler)
        elif isinstance(sampler, NestedSampler):
            self._init_from_jaxns(sampler)
        elif isinstance(sampler, ReactiveNestedSampler):
            self._init_from_ultranest(sampler)
        else:
            self._init_from_nautilus(sampler)

    def __repr__(self):
        tabs = self._tabs()
        return (
            f'Parameters\n{tabs["params"]}\n\n'
            f'Fit Statistics\n{tabs["stat"]}\n\n'
            f'Information Criterion\n{tabs["ic"]}\n\n'
            f'Pareto k diagnostic\n{tabs["k"]}\n'
        )

    def _repr_html_(self):
        """The repr in Jupyter notebook environment."""
        tabs = self._tabs()
        params_tab = tabs['params'].get_html_string(format=True)
        stat_tab = tabs['stat'].get_html_string(format=True)
        ic_tab = tabs['ic'].get_html_string(format=True)
        k_tab = tabs['k'].get_html_string(format=True)
        return (
            '<details open><summary><b>Posterior Result</b></summary>'
            '<details open style="padding-left: 1em">'
            f'<summary><b>Parameters</b></summary>{params_tab}</details>'
            '<details open style="padding-left: 1em">'
            f'<summary><b>Statistics</b></summary>{stat_tab}</details>'
            '<details open style="padding-left: 1em">'
            '<summary><b>Information Criterion</b></summary>'
            f'{ic_tab}</details>'
            '<details open style="padding-left: 1em">'
            f'<summary><b>Pareto k diagnostic</b></summary>{k_tab}</details>'
            '</details>'
        )

    def _tabs(self):
        if self._info_tabs is not None:
            return self._info_tabs
        params_name = self._helper.params_names['all']
        params = self.idata['posterior'][params_name]
        mean = params.mean()
        std = params.std(ddof=1)
        median = params.median()
        ci = params.quantile(0.5 + 0.683 * np.array([-0.5, 0.5])) - median
        ess = self.ess
        rhat = self.rhat
        rows = [
            [
                k,
                f'{mean[k]:.4g}',
                f'{std[k]:.4g}',
                f'{median[k]:.4g}',
                f'[{ci[k][0]:.4g}, {ci[k][1]:.4g}]',
                f'{ess[k]}',
                f'{rhat[k]:.2f}' if not np.isnan(rhat[k]) else 'N/A',
            ]
            for k in params_name
        ]
        names = [
            'Parameter',
            'Mean',
            'Std',
            'Median',
            '68.3% Quantile CI',
            'ESS',
            'Rhat',
        ]
        params_tab = make_pretty_table(names, rows)

        stat_type = self._helper.statistic
        deviance = self.deviance
        rows = [
            [
                i,
                stat_type[i],
                f'{deviance[i]["mean"]:.2f}',
                f'{deviance[i]["median"]:.2f}',
                j,
            ]
            for i, j in self.ndata.items()
            if i != 'total'
        ]
        rows.append(
            [
                'Total',
                'stat/dof',
                f'{deviance["total"]["mean"]:.2f}/{self.dof}',
                f'{deviance["total"]["median"]:.2f}/{self.dof}',
                self.ndata['total'],
            ]
        )
        names = [
            'Data',
            'Statistic',
            'Mean',
            'Median',
            'Channels',
        ]
        stat_tab = make_pretty_table(names, rows)

        loo = self.loo
        waic = self.waic
        rows = [
            [
                'LOOIC',
                f'{loo.elpd_loo:.2f} ± {loo.se:.2f}',
                f'{loo.p_loo:.2f}',
            ],
            [
                'WAIC',
                f'{waic.elpd_waic:.2f} ± {waic.se:.2f}',
                f'{waic.p_waic:.2f}',
            ],
        ]
        names = ['Method', 'Deviance', 'p']
        ic_tab = make_pretty_table(names, rows)

        ranges = ['(-Inf, 0.5]', '(0.5, 0.7]', '(0.7, 1]', '(1, Inf)']
        flags = ['good', 'ok', 'bad', 'very bad']
        bins = np.asarray([-np.inf, 0.5, 0.7, 1, np.inf])
        counts, *_ = np.histogram(loo.pareto_k.values, bins)
        pct = [f'{i:.1%}' for i in counts / np.sum(counts)]
        rows = list(zip(ranges, flags, counts, pct))
        names = ['Range', 'Flag', 'Count', 'Pct.']
        k_tab = make_pretty_table(names, rows)

        self._info_tabs = {
            'params': params_tab,
            'stat': stat_tab,
            'ic': ic_tab,
            'k': k_tab,
        }
        return self._info_tabs

    @property
    def plot(self) -> PosteriorResultPlotter:
        if self._plotter is None:
            self._plotter = PosteriorResultPlotter(self)
        return self._plotter

    def ci(
        self,
        params: str | Iterable[str] | None = None,
        cl: float | int = 1,
        hdi: bool = False,
    ) -> CredibleInterval:
        """Calculate credible intervals.

        Parameters
        ----------
        params : str or sequence of str, optional
            Parameters to calculate confidence intervals. If not specified,
            calculate for parameters of interest.
        cl : float or int, optional
            The credible level of samples within the credible interval. If
            0 < `cl` < 1, the value is interpreted as the probability mass.
            If `cl` >= 1, it is interpreted as the number of standard
            deviations. For example, ``cl=1`` produces a 1-sigma or 68.3%
            credible interval. The default is 1.
        hdi : bool, optional
            Whether to return the highest density interval. The default is
            False, which means an equal tailed interval is returned.

        Returns
        -------
        CredibleInterval
            The credible interval.
        """
        if cl <= 0.0:
            raise ValueError('cl must be non-negative')

        params = check_params(params, self._helper)

        cl_ = 1.0 - 2.0 * stats.norm.sf(cl) if cl >= 1.0 else cl

        if hdi:
            median = self.idata['posterior'].median()
            median = {
                k: float(v) for k, v in median.data_vars.items() if k in params
            }
            interval = az.hdi(self.idata, cl_, var_names=params)
            interval = {
                k: (float(v[0]), float(v[1]))
                for k, v in interval.data_vars.items()
            }
        else:
            q = [0.5, 0.5 - cl_ / 2.0, 0.5 + cl_ / 2.0]
            quantile = self.idata['posterior'].quantile(q)
            quantile = {
                k: v for k, v in quantile.data_vars.items() if k in params
            }
            median = {k: float(v[0]) for k, v in quantile.items()}
            interval = {
                k: (float(v[1]), float(v[2])) for k, v in quantile.items()
            }

        error = {
            k: (interval[k][0] - median[k], interval[k][1] - median[k])
            for k in params
        }

        return CredibleInterval(
            median=_format_result(median, params),
            intervals=_format_result(interval, params),
            errors=_format_result(error, params),
            cl=cl_,
            method='HDI' if hdi else 'ETI',
        )

    def _calc_flux(
        self,
        egrid: JAXArray,
        cl: float | int,
        hdi: bool,
        energy: bool,
        comps: bool,
        params: dict[str, JAXArray] | None,
    ) -> dict[str, Q | float]:
        if energy:
            unit = u.Unit('erg cm^-2 s^-1')
        else:
            unit = u.Unit('ph cm^-2 s^-1')

        post = self._params_dist
        n = [i.size for i in post.values()][0]
        if params is not None:
            params = dict(params)
            params = {k: jnp.full(n, v) for k, v in params.items()}
            post = post | params
        devices = create_device_mesh((jax.local_device_count(),))
        mesh = Mesh(devices, axis_names=('i',))
        p = PartitionSpec()
        pi = PartitionSpec('i')
        fn = shard_map(
            f=self._flux_fn,
            mesh=mesh,
            in_specs=(p, pi, p, p),
            out_specs=pi,
            check_rep=False,
        )
        flux = jax.device_get(fn(egrid, post, energy, comps))
        cl_ = 1.0 - 2.0 * stats.norm.sf(cl) if cl >= 1.0 else cl
        if hdi:
            ci_fn = lambda x: az.hdi(x, cl_)
        else:
            q = 0.5 + np.array([-0.5, 0.5]) * cl_
            ci_fn = lambda x: np.quantile(x, q)
        median = jax.tree_map(lambda x: np.median(x), flux)
        intervals = jax.tree_map(ci_fn, flux)
        errors = jax.tree_map(
            lambda x, y: (y[0] - x, y[1] - x), median, intervals
        )
        add_unit = lambda x: x * unit
        return {
            'median': jax.tree_map(add_unit, median),
            'intervals': jax.tree_map(add_unit, intervals),
            'errors': jax.tree_map(add_unit, errors),
            'cl': cl_,
            'dist': jax.tree_map(add_unit, flux),
            'n': n,
        }

    def flux(
        self,
        emin: float | int,
        emax: float | int,
        cl: float | int = 1,
        energy: bool = True,
        ngrid: int = 1000,
        hdi: bool = False,
        comps: bool = False,
        log: bool = True,
        params: dict[str, float | int] | None = None,
    ) -> PosteriorFlux:
        r"""Calculate the flux of model.

        .. warning::
            The flux is calculated by trapezoidal rule, and is accurate only
            if enough numbers of energy grids are used.

        Parameters
        ----------
        emin : float or int
            Minimum value of energy range, in units of keV.
        emax : float or int
            Maximum value of energy range, in units of keV.
        cl : float or int, optional
            The credible level of samples within the credible interval. If
            0 < `cl` < 1, the value is interpreted as the probability mass.
            If `cl` >= 1, it is interpreted as the number of standard
            deviations. For example, ``cl=1`` produces a 1-sigma or 68.3%
            credible interval. The default is 1.
        energy : bool, optional
            When True, calculate energy flux in units of erg cm⁻² s⁻¹;
            otherwise calculate photon flux in units of ph cm⁻² s⁻¹.
            The default is True.
        ngrid : int, optional
            The energy grid number to use in integration. The default is 1000.

        Other Parameters
        ----------------
        hdi : bool, optional
            Whether to return the highest density interval. The default is
            False, which means an equal tailed interval is returned.
        comps : bool, optional
            Whether to return the result of each component. The default is
            False.
        log : bool, optional
            Whether to use logarithmically regular energy grid. The default is
            True.
        params : dict, optional
            Parameters dict to overwrite the fitted parameters.

        Returns
        -------
        PosteriorFlux
            The flux of the model.
        """
        if log:
            egrid = jnp.geomspace(emin, emax, ngrid)
        else:
            egrid = jnp.linspace(emin, emax, ngrid)

        flux = self._calc_flux(egrid, cl, hdi, energy, comps, params)

        return PosteriorFlux(
            median=flux['median'],
            intervals=flux['intervals'],
            errors=flux['errors'],
            cl=flux['cl'],
            dist=flux['dist'],
            energy=bool(energy),
            n=flux['n'],
        )

    def lumin(
        self,
        emin_rest: float | int,
        emax_rest: float | int,
        z: float | int,
        cl: float | int = 1,
        ngrid: int = 1000,
        hdi: bool = False,
        comps: bool = False,
        log: bool = True,
        params: dict[str, float | int] | None = None,
        cosmo: LambdaCDM = Planck18,
    ) -> PosteriorLumin:
        """Calculate the luminosity of model.

        .. warning::
            The luminosity is calculated by trapezoidal rule, and is accurate
            only if enough numbers of energy grids are used.

        Parameters
        ----------
        emin_rest : float or int
            Minimum value of rest-frame energy range, in units of keV.
        emax_rest : float or int
            Maximum value of rest-frame energy range, in units of keV.
        z : float or int
            Redshift of the source.
        cl : float or int, optional
            The credible level of samples within the credible interval. If
            0 < `cl` < 1, the value is interpreted as the probability mass.
            If `cl` >= 1, it is interpreted as the number of standard
            deviations. For example, ``cl=1`` produces a 1-sigma or 68.3%
            credible interval. The default is 1.
        ngrid : int, optional
            The energy grid number to use in integration. The default is 1000.

        Other Parameters
        ----------------
        hdi : bool, optional
            Whether to return the highest density interval. The default is
            False, which means an equal tailed interval is returned.
        comps : bool, optional
            Whether to return the result of each component. The default is
            False.
        log : bool, optional
            Whether to use logarithmically regular energy grid. The default is
            True.
        params : dict, optional
            Parameters dict to overwrite the fitted parameters.
        cosmo : LambdaCDM, optional
            Cosmology model used to calculate luminosity. The default is
            Planck18.

        Returns
        -------
        PosteriorLumin
            The luminosity of the model.
        """
        if log:
            egrid = jnp.geomspace(emin_rest, emax_rest, ngrid) / (1.0 + z)
        else:
            egrid = jnp.linspace(emin_rest, emax_rest, ngrid) / (1.0 + z)

        z = float(z)
        flux = self._calc_flux(egrid, cl, hdi, True, comps, params)

        factor = 4.0 * np.pi * cosmo.luminosity_distance(z) ** 2
        to_lumin = lambda x: (x * factor).to('erg s^-1')

        return PosteriorLumin(
            median=jax.tree_map(to_lumin, flux['median']),
            intervals=jax.tree_map(to_lumin, flux['intervals']),
            errors=jax.tree_map(to_lumin, flux['errors']),
            cl=flux['cl'],
            dist=jax.tree_map(to_lumin, flux['dist']),
            n=flux['n'],
            z=z,
            cosmo=cosmo,
        )

    def eiso(
        self,
        emin_rest: float | int,
        emax_rest: float | int,
        z: float | int,
        duration: float | int,
        cl: float | int = 1,
        ngrid: int = 1000,
        hdi: bool = False,
        comps: bool = False,
        log: bool = True,
        params: dict[str, float | int] | None = None,
        cosmo: LambdaCDM = Planck18,
    ) -> PosteriorEIso:
        r"""Calculate the isotropic emission energy of model.

        .. warning::
            The :math:`E_\mathrm{iso}` is calculated by trapezoidal rule,
            and is accurate only if enough numbers of energy grids are used.

        Parameters
        ----------
        emin_rest : float or int
            Minimum value of rest-frame energy range, in units of keV.
        emax_rest : float or int
            Maximum value of rest-frame energy range, in units of keV.
        z : float or int
            Redshift of the source.
        duration : float or int
            Observed duration of the source, in units of seconds.
        cl : float or int, optional
            The credible level of samples within the credible interval. If
            0 < `cl` < 1, the value is interpreted as the probability mass.
            If `cl` >= 1, it is interpreted as the number of standard
            deviations. For example, ``cl=1`` produces a 1-sigma or 68.3%
            credible interval. The default is 1.
        ngrid : int, optional
            The energy grid number to use in integration. The default is
            1000.

        Other Parameters
        ----------------
        hdi : bool, optional
            Whether to return the highest density interval. The default is
            False, which means an equal tailed interval is returned.
        comps : bool, optional
            Whether to return the result of each component. The default is
            False.
        log : bool, optional
            Whether to use logarithmically regular energy grid. The default
            is True.
        params : dict, optional
            Parameters dict to overwrite the fitted parameters.
        cosmo : LambdaCDM, optional
            Cosmology model used to calculate luminosity. The default is
            Planck18.

        Returns
        -------
        PosteriorEIso
            The isotropic emission energy of the model.
        """
        lumin = self.lumin(
            emin_rest, emax_rest, z, cl, ngrid, hdi, comps, log, params, cosmo
        )

        # This includes correction for energy redshift and time dilation.
        factor = duration / (1 + z) * u.s
        to_eiso = lambda x: (x * factor).to('erg')

        return PosteriorEIso(
            median=jax.tree_map(to_eiso, lumin.median),
            intervals=jax.tree_map(to_eiso, lumin.intervals),
            errors=jax.tree_map(to_eiso, lumin.errors),
            cl=lumin.cl,
            dist=jax.tree_map(to_eiso, lumin.dist),
            n=lumin.n,
            z=lumin.z,
            duration=float(duration),
            cosmo=lumin.cosmo,
        )

    def ppc(
        self,
        n: int = 10000,
        seed: int | None = None,
        parallel: bool = True,
        progress: bool = True,
        update_rate: int = 50,
    ):
        """Perform posterior predictive check.

        Parameters
        ----------
        n : int, optional
            The number of posterior predictions. The default is 10000.
        seed : int, optional
            The seed of random number generator used in posterior predictions.
        parallel : bool, optional
            Whether to run simulation fit in parallel. The default is True.
        progress : bool, optional
            Whether to display progress bar. The default is True.
        update_rate : int, optional
            The update rate of progress bar. The default is 50.
        """
        # reuse the previous result if all setup is the same
        if self._ppc and self._ppc.n == n and self._ppc.seed == seed:
            return

        helper = self._helper
        free_params = helper.params_names['free']
        n = int(n)
        seed = helper.seed['pred'] if seed is None else int(seed)

        # randomly select n samples from posterior
        rng = np.random.default_rng(seed)
        idata = self.idata
        i = rng.integers(0, idata['posterior'].chain.size, n)
        j = rng.integers(0, idata['posterior'].draw.size, n)
        params = {
            k: v.values[i, j]
            for k, v in idata['posterior'][free_params].data_vars.items()
        }
        models = {
            k: v.values[i, j]
            for k, v in helper.get_models(idata['posterior']).items()
        }

        # perform ppc
        result = helper.simulate_and_fit(
            seed, params, models, 1, parallel, progress, update_rate, 'PPC'
        )
        valid = result.pop('valid')
        result = jax.tree_map(lambda x: x[valid], result)

        self._ppc = PPCResult(
            params_rep=params,
            models_rep=models,
            data=result['data'],
            params_fit=result['params'],
            models_fit=result['models'],
            deviance=result['deviance'],
            p_value=jax.tree_map(
                lambda obs, sim: np.sum(sim >= obs, axis=0) / len(sim),
                self._mle['deviance'],
                result['deviance'],
            ),
            n=n,
            n_valid=np.sum(valid),
            seed=seed,
        )

    @property
    def gof(self) -> dict[str, float]:
        if self._ppc is None:
            raise RuntimeError(
                'PosteriorResult.ppc() must be called to assess gof'
            )
        p_value = self._ppc.p_value
        p_value = p_value['group'] | {'total': p_value['total']}
        return {k: float(p_value[k]) for k in self.ndata.keys()}

    @property
    def _mle(self):
        """MLE result."""
        if self._mle_result is None:
            mle_result = {}
            helper = self._helper

            # MLE information of the model
            free_params = helper.params_names['free']
            mle_idx = self.idata['log_likelihood']['total'].argmax(...)
            init = self.idata['posterior'][free_params].sel(**mle_idx)
            init = {k: v.values for k, v in init.data_vars.items()}
            init = helper.constr_dic_to_unconstr_arr(init)
            mle_unconstr = self._fit._optimize_lm(init, throw=False)[0]

            # MLE of model params in constrained space
            mle, cov = jax.device_get(helper.get_mle(mle_unconstr))
            err = np.sqrt(np.diagonal(cov))
            params_names = helper.params_names['all']
            mle_result['params'] = dict(zip(params_names, zip(mle, err)))

            sites = jax.device_get(jax.jit(helper.get_sites)(mle_unconstr))

            # model deviance at MLE
            loglike = sites['loglike']
            # drop unnecessary terms
            loglike.pop('data')
            loglike.pop('channels')
            mle_result['deviance'] = jax.tree_map(lambda x: -2.0 * x, loglike)

            # model values at MLE
            mle_result['models'] = sites['models']

            self._mle_result = mle_result
        return self._mle_result

    @property
    def mle(self) -> dict[str, tuple[float, float]]:
        """MLE parameters."""
        return dict(self._mle['params'])

    @property
    def idata(self) -> az.InferenceData:
        """ArviZ InferenceData."""
        return self._idata

    @property
    def _params_dist(self) -> dict[str, JAXArray]:
        """Posterior parameters, the size is truncated to <= nmax."""
        if self._params is not None:
            return self._params

        nmax = 10000
        post = self.idata['posterior'][self._helper.params_names['free']]
        n = post.chain.size * post.draw.size
        if n > nmax:
            n = nmax - nmax % jax.local_device_count()
            rng = np.random.default_rng(self._helper.seed['mcmc'])
            i = rng.integers(0, post.chain.size, n)
            j = rng.integers(0, post.draw.size, n)
            post = {k: v.values[i, j] for k, v in post.items()}
        else:
            post = {k: np.hstack(v.values) for k, v in post.data_vars.items()}
        self._params = post
        return self._params

    def _init_from_numpyro(self, sampler: MCMC):
        helper = self._helper
        params_names = helper.params_names
        samples = sampler.get_samples(group_by_chain=True)

        # stats of samples
        rename = {'num_steps': 'n_steps'}
        sample_stats = {}
        for k, v in sampler.get_extra_fields(group_by_chain=True).items():
            name = rename.get(k, k)
            value = jax.device_get(v).copy()
            sample_stats[name] = value
            if k == 'num_steps':
                sample_stats['tree_depth'] = np.log2(value).astype(int) + 1

        # attrs for each group of arviz.InferenceData
        attrs = {
            'elisa_version': __version__,
            'inference_library': 'numpyro',
            'inference_library_version': numpyro.__version__,
        }

        # create arviz.InferenceData
        self._generate_idata(samples, attrs, sample_stats)

        # effective sample size
        params = helper.get_params(samples)
        ess = az.ess(params)
        self._ess = {k: int(ess[k].values) for k in params.keys()}

        # relative mcmc efficiency
        # the calculation of reff is according to arviz loo:
        # https://github.com/arviz-devs/arviz/blob/main/arviz/stats/stats.py#L770
        if sampler.num_chains == 1:
            self._reff = 1.0
        else:
            # use only free parameters to calculate reff
            free = {k: params[k] for k in params_names['free']}
            reff_p = az.ess(free, method='mean', relative=True)
            self._reff = np.hstack(list(reff_p.data_vars.values())).mean()

        # model evidence
        self._lnZ = (None, None)

    def _init_from_jaxns(self, sampler: NestedSampler):
        helper = self._helper
        result = sampler._results

        # get posterior samples
        total = result.total_num_samples
        rng_key = jax.random.PRNGKey(helper.seed['mcmc'])
        samples = jax.tree_map(
            lambda x: x[None, ...],
            sampler.get_samples(rng_key, total),
        )

        # attrs for each group of arviz.InferenceData
        attrs = {
            'elisa_version': __version__,
            'inference_library': 'jaxns',
            'inference_library_version': metadata.version('jaxns'),
        }

        self._generate_idata(samples, attrs)

        # effective sample size
        ess = int(result.ESS)
        self._ess = {p: ess for p in self._helper.params_names['all']}
        # relative mcmc efficiency
        self._reff = float(ess / result.total_num_samples)
        # model evidence
        self._lnZ = (float(result.log_Z_mean), float(result.log_Z_uncert))

    def _init_from_ultranest(self, sampler: ReactiveNestedSampler):
        result = sampler._transform_back(sampler.results['samples'])
        nsamples = len(sampler.results['samples'])
        ncores = jax.local_device_count()
        ndrop = nsamples % ncores

        # get posterior samples
        samples = jax.tree_map(lambda x: x[None, : nsamples - ndrop], result)

        # attrs for each group of arviz.InferenceData
        attrs = {
            'elisa_version': __version__,
            'inference_library': 'ultranest',
            'inference_library_version': ultranest.__version__,
        }

        self._generate_idata(samples, attrs)

        # effective sample size
        ess = int(sampler.results['ess'])
        self._ess = {p: ess for p in self._helper.params_names['all']}
        # relative mcmc efficiency
        self._reff = float(ess / nsamples)
        # model evidence
        self._lnZ = (
            float(sampler.results['logz']),
            float(sampler.results['logzerr']),
        )

    def _init_from_nautilus(self, sampler: Sampler):
        result = sampler.posterior(equal_weight=True)[0]
        result = sampler._transform_back(result)
        ncores = jax.local_device_count()

        # get posterior samples
        samples = jax.tree_map(
            lambda x: x[None, : len(x) - len(x) % ncores],
            result,
        )

        # attrs for each group of arviz.InferenceData
        attrs = {
            'elisa_version': __version__,
            'inference_library': 'nautilus',
            'inference_library_version': nautilus.__version__,
        }

        self._generate_idata(samples, attrs)

        # effective sample size
        ess = int(sampler.n_eff)
        self._ess = {p: ess for p in self._helper.params_names['all']}
        # relative mcmc efficiency
        total_sample = len(sampler.posterior(equal_weight=False)[0])
        self._reff = float(ess / total_sample)
        # model evidence
        self._lnZ = (float(sampler.log_z), None)

    def _generate_idata(self, samples, attrs, sample_stats=None):
        samples = jax.tree_map(jax.device_get, samples)
        helper = self._helper

        params = helper.get_params(samples)
        models = helper.get_models(samples)
        posterior = params | models
        posterior_predictive = helper.simulate(helper.seed['pred'], models, 1)
        loglike = helper.get_loglike(samples)
        group = {f'{k}_total': v for k, v in loglike['group'].items()}
        loglike = (
            loglike['data']
            | loglike['point']
            | group
            | {'channels': loglike['channels']}
            | {'total': loglike['total']}
        )

        # get observation counts data
        obs_data = helper.obs_data

        # coords and dims of arviz.InferenceData
        coords = dict(helper.channels)

        dims = {'channels': ['channel']}
        for i in helper.data_names:
            dim = [f'{i}_channel']
            dims[i] = dims[f'{i}_Non'] = dims[f'{i}_Non_model'] = dim

            if f'{i}_Noff' in obs_data:
                dims[f'{i}_Noff'] = dims[f'{i}_Noff_model'] = dim

        # create InferenceData
        self._idata = az.from_dict(
            posterior=posterior,
            posterior_predictive=posterior_predictive,
            sample_stats=sample_stats,
            log_likelihood=loglike,
            observed_data=obs_data,
            coords=coords,
            dims=dims,
            posterior_attrs=attrs,
            posterior_predictive_attrs=attrs,
            sample_stats_attrs=attrs,
            log_likelihood_attrs=attrs,
            observed_data_attrs=attrs,
        )

    @property
    def deviance(self) -> dict:
        """Mean and median of model deviance."""
        if self._deviance is None:
            stat_keys = {
                i: f'{i}_total' if i != 'total' else i
                for i in self.ndata.keys()
            }
            keys = list(stat_keys.values())
            deviance = -2.0 * self.idata['log_likelihood'][keys]
            deviance_mean = deviance.mean()
            deviance_median = deviance.median()
            self._deviance = {
                k: {
                    'mean': float(deviance_mean[v]),
                    'median': float(deviance_median[v]),
                }
                for k, v in stat_keys.items()
            }

        return self._deviance

    @property
    def reff(self) -> float:
        """Relative MCMC efficiency."""
        return self._reff

    @property
    def ess(self) -> dict[str, int]:
        """Effective MCMC sample size."""
        return self._ess

    @property
    def rhat(self) -> dict[str, float]:
        """Computes split R-hat over MCMC chains.

        In general, only fully trust the sample if R-hat is less than 1.01. In
        the early workflow, R-hat below 1.1 is often sufficient. See [1]_ for
        more information.

        References
        ----------
        .. [1] https://arxiv.org/abs/1903.08008
        """
        if self._rhat is None:
            params_names = self._helper.params_names['all']
            posterior = self.idata['posterior'][params_names]

            if len(posterior['chain']) == 1:
                rhat = {k: float('nan') for k in posterior.data_vars.keys()}
            else:
                rhat = {
                    k: float(v.values)
                    for k, v in az.rhat(posterior).data_vars.items()
                }

            self._rhat = rhat

        return self._rhat

    @property
    def divergence(self) -> int:
        """Number of divergent samples."""
        if self._divergence is None:
            if 'sample_stats' in self.idata:
                n = int(self.idata['sample_stats']['diverging'].sum())
            else:
                n = 0

            self._divergence = n

        return self._divergence

    @property
    def waic(self) -> ELPDData:
        """The widely applicable information criterion (WAIC).

        Estimates the expected log point-wise predictive density (elpd) using
        WAIC. Also calculates the WAIC's standard error and the effective
        number of parameters. See [1]_ and [2]_ for more information.

        References
        ----------
        .. [1] https://arxiv.org/abs/1507.04544
        .. [2] https://arxiv.org/abs/1004.2316
        """
        if self._waic is None:
            self._waic = az.waic(
                self.idata, var_name='channels', scale='deviance'
            )

        return self._waic

    @property
    def loo(self) -> ELPDData:
        """Pareto-smoothed importance sampling leave-one-out cross-validation
        (PSIS-LOO-CV).

        Estimates the expected log point-wise predictive density (elpd) using
        PSIS-LOO-CV. Also calculates LOO's standard error and the effective
        number of parameters. For more information, see [1]_, [2]_ and [3]_.

        References
        ----------
        .. [1] https://avehtari.github.io/modelselection/CV-FAQ.html
        .. [2] https://arxiv.org/abs/1507.04544
        .. [3] https://arxiv.org/abs/1507.02646
        """
        if self._loo is None:
            self._loo = az.loo(
                self.idata,
                var_name='channels',
                reff=self.reff,
                scale='deviance',
            )

        return self._loo

    @property
    def lnZ(self) -> tuple[float, float] | tuple[None, None]:
        """Log model evidence and uncertainty."""
        return self._lnZ

    @property
    def _loo_pit(self) -> dict[str, tuple]:
        """Leave-one-out probability integral transform."""
        if self._pit is not None:
            return self._pit

        idata = self.idata
        reff = self.reff
        helper = self._helper
        stack_kwargs = {'__sample__': ('chain', 'draw')}
        log_weights, kss = az.psislw(
            -idata['log_likelihood']['channels'].stack(**stack_kwargs), reff
        )
        y_hat = idata['posterior_predictive']['channels'].stack(**stack_kwargs)
        loo_pit = az.loo_pit(
            y=idata['observed_data']['channels'],
            y_hat=y_hat,
            log_weights=log_weights,
        )

        loo_pit = {
            name: loo_pit.sel(channel=data.channel).values
            for name, data in helper.data.items()
        }

        discrete_stats = {'cstat', 'pstat', 'wstat'}
        data_stats = helper.statistic
        has_discrete = discrete_stats.intersection(data_stats.values())
        if has_discrete:
            data_minus = {}
            for k, d in helper.data.items():
                unit = 1.0 / (d.channel_width * d.spec_exposure)
                if data_stats[k] in {'cstat', 'pstat'}:
                    data_minus[k] = (d.spec_counts - 1.0) * unit
                elif data_stats[k] == 'wstat':
                    # Get the next small net spectrum values
                    data_minus[k] = (
                        np.maximum(
                            (d.spec_counts - 1.0)
                            - d.back_ratio * d.back_counts,
                            d.spec_counts
                            - d.back_ratio * (d.back_counts + 1.0),
                        )
                        * unit
                    )
                else:  # chi2, pgstat
                    data_minus[k] = d.ce

            y_miuns = idata['observed_data']['channels'].copy()
            y_miuns.data = np.hstack(list(data_minus.values()))

            loo_pit_minus = az.loo_pit(
                y=y_miuns,
                y_hat=y_hat,
                log_weights=log_weights,
            )
            loo_pit_minus = {
                name: loo_pit_minus.sel(channel=data.channel).values
                for name, data in helper.data.items()
            }
        else:
            loo_pit_minus = loo_pit

        self._pit = {
            name: (loo_pit_minus[name], loo_pit[name])
            for name in loo_pit.keys()
        }
        return self._pit


class ConfidenceInterval(NamedTuple):
    """Confidence interval result."""

    mle: dict[str, float]
    """MLE of the model parameters."""

    intervals: dict[str, tuple[float, float]]
    """The confidence intervals."""

    errors: dict[str, tuple[float, float]]
    """The confidence intervals in error form."""

    cl: float
    """The confidence level."""

    method: str
    """Method used to calculate the confidence interval."""

    status: dict
    """Status of the calculation progress."""


class MLEFlux(NamedTuple):
    """The flux of the MLE model."""

    mle: dict[str, Q] | dict[str, dict[str, Q]]
    """The model flux at MLE."""

    intervals: dict[str, tuple[Q, Q]] | dict[str, dict[str, tuple[Q, Q]]]
    """The confidence intervals of the model flux."""

    errors: dict[str, tuple[Q, Q]] | dict[str, dict[str, tuple[Q, Q]]]
    """The confidence intervals of the model flux in error form."""

    dist: dict[str, Q] | dict[str, dict[str, Q]]
    """Bootstrap flux distribution."""

    cl: float
    """The confidence level."""

    energy: bool
    """Whether the flux is in energy flux. False for photon flux."""

    n: int
    """Numbers of bootstrap samples."""


class MLELumin(NamedTuple):
    """The luminosity of the MLE model."""

    mle: dict[str, Q] | dict[str, dict[str, Q]]
    """The model luminosity at MLE."""

    intervals: dict[str, tuple[Q, Q]] | dict[str, dict[str, tuple[Q, Q]]]
    """The confidence intervals of the model luminosity."""

    errors: dict[str, tuple[Q, Q]] | dict[str, dict[str, tuple[Q, Q]]]
    """The confidence intervals of the model luminosity in error form."""

    dist: dict[str, Q] | dict[str, dict[str, Q]]
    """Bootstrap luminosity distribution."""

    cl: float
    """The confidence level."""

    n: int
    """Numbers of bootstrap samples."""

    z: float
    """Redshift of the source."""

    cosmo: LambdaCDM
    """Cosmology model used to calculate luminosity."""


class MLEEIso(NamedTuple):
    """The isotropic emission energy of the MLE model."""

    mle: dict[str, Q] | dict[str, dict[str, Q]]
    r"""The model Eiso at MLE."""

    intervals: dict[str, tuple[Q, Q]] | dict[str, dict[str, tuple[Q, Q]]]
    """The confidence intervals of the model Eiso."""

    errors: dict[str, tuple[Q, Q]] | dict[str, dict[str, tuple[Q, Q]]]
    """The confidence intervals of the model Eiso in error form."""

    dist: dict[str, Q] | dict[str, dict[str, Q]]
    """Bootstrap Eiso distribution."""

    cl: float
    """The confidence level."""

    n: int
    """Numbers of bootstrap samples."""

    z: float
    """Redshift of the source."""

    duration: float
    """Observed duration of the source."""

    cosmo: LambdaCDM
    """Cosmology model used to calculate Eiso."""


class BootstrapResult(NamedTuple):
    """Parametric bootstrap result."""

    mle: dict
    """MLE of the model parameters."""

    data: dict
    """Simulation data based on MLE."""

    models: dict
    """Bootstrap models."""

    params: dict
    """Bootstrap parameters."""

    deviance: dict
    """Bootstrap deviance."""

    p_value: dict
    """Model fitness :math:`p`-value."""

    n: int
    """Numbers of bootstrap."""

    n_valid: int
    """Numbers of valid bootstrap."""

    seed: int
    """Seed of random number generator used in simulation."""


class CredibleInterval(NamedTuple):
    """Credible interval result."""

    median: dict[str, float]
    """Median of the model parameters."""

    intervals: dict[str, tuple[float, float]]
    """The credible intervals."""

    errors: dict[str, tuple[float, float]]
    """The credible intervals in error form."""

    cl: float
    """The credible level."""

    method: str
    """Highest Density Interval (HDI), or equal tailed interval (ETI)."""


class PosteriorFlux(NamedTuple):
    """Posterior flux."""

    median: dict[str, Q] | dict[str, dict[str, Q]]
    """The median flux."""

    intervals: dict[str, tuple[Q, Q]] | dict[str, dict[str, tuple[Q, Q]]]
    """The credible intervals of the model flux."""

    errors: dict[str, tuple[Q, Q]] | dict[str, dict[str, tuple[Q, Q]]]
    """The credible intervals of the model flux in error form."""

    dist: dict[str, Q] | dict[str, dict[str, Q]]
    """Posterior flux distribution."""

    cl: float
    """The credible level."""

    energy: bool
    """Whether the flux is in energy flux. False for photon flux."""

    n: int
    """Numbers of posterior samples."""


class PosteriorLumin(NamedTuple):
    """Posterior luminosity."""

    median: dict[str, Q] | dict[str, dict[str, Q]]
    """The median luminosity."""

    intervals: dict[str, tuple[Q, Q]] | dict[str, dict[str, tuple[Q, Q]]]
    """The credible intervals of the model luminosity."""

    errors: dict[str, tuple[Q, Q]] | dict[str, dict[str, tuple[Q, Q]]]
    """The credible intervals of the model luminosity in error form."""

    dist: dict[str, Q] | dict[str, dict[str, Q]]
    """Posterior distribution of luminosity."""

    cl: float
    """The credible level."""

    n: int
    """Numbers of posterior samples."""

    z: float
    """Redshift of the source."""

    cosmo: LambdaCDM
    """Cosmology model used to calculate luminosity."""


class PosteriorEIso(NamedTuple):
    """Posterior isotropic emission energy."""

    median: dict[str, Q] | dict[str, dict[str, Q]]
    r"""The median Eiso."""

    intervals: dict[str, tuple[Q, Q]] | dict[str, dict[str, tuple[Q, Q]]]
    """The credible intervals of the model Eiso."""

    errors: dict[str, tuple[Q, Q]] | dict[str, dict[str, tuple[Q, Q]]]
    """The credible intervals of the model Eiso in error form."""

    dist: dict[str, Q] | dict[str, dict[str, Q]]
    """Posterior distribution of Eiso."""

    cl: float
    """The credible level."""

    n: int
    """Numbers of posterior samples."""

    z: float
    """Redshift of the source."""

    duration: float
    """Observed duration of the source."""

    cosmo: LambdaCDM
    """Cosmology model used to calculate Eiso."""


class PPCResult(NamedTuple):
    """Posterior predictive check result."""

    params_rep: dict
    """Posterior of free parameters used to perform ppc."""

    models_rep: dict
    """Models' values corresponding to `params_rep`."""

    data: dict
    """Posterior predictive data."""

    params_fit: dict
    """Best fit parameters of posterior predictive data."""

    deviance: dict
    """Deviance of posterior predictive data and best fit models."""

    models_fit: dict
    """Best fit models' values of posterior predictive data."""

    p_value: dict
    """Posterior predictive :math:`p`-value."""

    n: int
    """Numbers of posterior prediction."""

    n_valid: int
    """Numbers of valid ppc."""

    seed: int
    """Seed of random number generator used in simulation."""


def _format_result(result: dict, order: Sequence[str]) -> dict:
    """Sort the result and use float type."""
    formatted = jax.tree_map(float, result)
    return {k: formatted[k] for k in order}
